# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from abc import ABC
from typing import Tuple, Optional

import fsspec
import rasterio
import rioxarray
import s3fs
import xarray as xr
import zarr
from rasterio.session import AWSSession

from xcube.core.zarrstore import LoggingZarrStore
# Note, we need the following reference to register the
# xarray property accessor
# noinspection PyUnresolvedReferences
from xcube.core.zarrstore import ZarrStoreHolder
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.fspath import is_local_fs
from xcube.util.jsonencoder import to_json_value
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from xcube.util.temp import new_temp_file
from ..accessor import FsDataAccessor
from ...datatype import DATASET_TYPE
from ...datatype import DataType
from ...error import DataStoreError

ZARR_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        log_access=JsonBooleanSchema(
            default=False
        ),
        cache_size=JsonIntegerSchema(
            minimum=0,
        ),
        group=JsonStringSchema(
            description='Group path.'
                        ' (a.k.a. path in zarr terminology.).',
            min_length=1,
        ),
        chunks=JsonObjectSchema(
            description='Optional chunk sizes along each dimension.'
                        ' Chunk size values may be None, "auto"'
                        ' or an integer value.',
            examples=[{'time': None, 'lat': 'auto', 'lon': 90},
                      {'time': 1, 'y': 512, 'x': 512}],
            additional_properties=True,
        ),
        decode_cf=JsonBooleanSchema(
            description='Whether to decode these variables,'
                        ' assuming they were saved according to'
                        ' CF conventions.',
            default=True,
        ),
        mask_and_scale=JsonBooleanSchema(
            description='If True, replace array values equal'
                        ' to attribute "_FillValue" with NaN. '
                        ' Use "scale_factor" and "add_offset"'
                        ' attributes to compute actual values.',
            default=True,
        ),
        decode_times=JsonBooleanSchema(
            description='If True, decode times encoded in the'
                        ' standard NetCDF datetime format '
                        'into datetime objects. Otherwise,'
                        ' leave them encoded as numbers.',
            default=True,
        ),
        decode_coords=JsonBooleanSchema(
            description='If True, decode the \"coordinates\"'
                        ' attribute to identify coordinates in '
                        'the resulting dataset.',
            default=True,
        ),
        drop_variables=JsonArraySchema(
            items=JsonStringSchema(min_length=1),
        ),
        consolidated=JsonBooleanSchema(
            description='Whether to open the store using'
                        ' Zarr\'s consolidated metadata '
                        'capability. Only works for stores that'
                        ' have already been consolidated.',
            default=False,
        ),
    ),
    required=[],
    additional_properties=False
)

ZARR_WRITE_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        group=JsonStringSchema(
            description='Group path.'
                        ' (a.k.a. path in zarr terminology.).',
            min_length=1,
        ),
        encoding=JsonObjectSchema(
            description='Nested dictionary with variable'
                        ' names as keys and dictionaries of'
                        ' variable specific encodings as values.',
            examples=[{
                'my_variable': {
                    'dtype': 'int16',
                    'scale_factor': 0.1
                }
            }],
            additional_properties=True,
        ),
        consolidated=JsonBooleanSchema(
            description='If True (the default), consolidate all metadata'
                        ' files ("**/.zarray", "**/.zattrs")'
                        ' into a single top-level file ".zmetadata"',
            default=True,
        ),
        append_dim=JsonStringSchema(
            description='If set, the dimension on which the'
                        ' data will be appended.',
            min_length=1,
        ),
    ),
    additional_properties=False
)


class DatasetFsDataAccessor(FsDataAccessor, ABC):
    """
    Opener/writer extension name: "dataset:<format>:<protocol>"
    """

    @classmethod
    def get_data_type(cls) -> DataType:
        return DATASET_TYPE


class DatasetZarrFsDataAccessor(DatasetFsDataAccessor):
    """
    Opener/writer extension name: "dataset:zarr:<protocol>"
    """

    @classmethod
    def get_format_id(cls) -> str:
        return 'zarr'

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_open_data_params_schema(self, data_id: str = None) \
            -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(
            ZARR_OPEN_DATA_PARAMS_SCHEMA
        )

    def open_data(self,
                  data_id: str,
                  **open_params) -> xr.Dataset:
        assert_instance(data_id, str, name='data_id')
        fs, root, open_params = self.load_fs(open_params)
        zarr_store = fs.get_mapper(data_id)
        cache_size = open_params.pop('cache_size', None)
        if isinstance(cache_size, int) and cache_size > 0:
            zarr_store = zarr.LRUStoreCache(zarr_store, max_size=cache_size)
        log_access = open_params.pop('log_access', None)
        if log_access:
            zarr_store = LoggingZarrStore(zarr_store,
                                          name=f'zarr_store({data_id!r})')
        consolidated = open_params.pop('consolidated',
                                       fs.exists(f'{data_id}/.zmetadata'))
        try:
            dataset = xr.open_zarr(zarr_store,
                                   consolidated=consolidated,
                                   **open_params)
        except ValueError as e:
            raise DataStoreError(f'Failed to open'
                                 f' dataset {data_id!r}: {e}') from e

        dataset.zarr_store.set(zarr_store)
        return dataset

    # noinspection PyMethodMayBeStatic
    def get_write_data_params_schema(self) -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(
            ZARR_WRITE_DATA_PARAMS_SCHEMA
        )

    def write_data(self,
                   data: xr.Dataset,
                   data_id: str,
                   replace=False,
                   **write_params) -> str:
        assert_instance(data, xr.Dataset, name='data')
        assert_instance(data_id, str, name='data_id')
        fs, root, write_params = self.load_fs(write_params)
        zarr_store = fs.get_mapper(data_id, create=True)
        log_access = write_params.pop('log_access', None)
        if log_access:
            zarr_store = LoggingZarrStore(zarr_store,
                                          name=f'zarr_store({data_id!r})')
        consolidated = write_params.pop('consolidated', True)
        try:
            data.to_zarr(zarr_store,
                         mode='w' if replace else None,
                         consolidated=consolidated,
                         **write_params)
        except ValueError as e:
            raise DataStoreError(f'Failed to write'
                                 f' dataset {data_id!r}: {e}') from e
        return data_id

    def delete_data(self,
                    data_id: str,
                    **delete_params):
        fs, root, delete_params = self.load_fs(delete_params)
        delete_params.pop('recursive', None)
        fs.delete(data_id, recursive=True, **delete_params)


NETCDF_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        # TODO: add more from xr.open_dataset()
    ),
    additional_properties=True,
)

NETCDF_WRITE_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        # TODO: add more from ds.to_netcdf()
    ),
    additional_properties=True,
)


class DatasetNetcdfFsDataAccessor(DatasetFsDataAccessor):
    """
    Opener/writer extension name: "dataset:netcdf:<protocol>"
    """

    @classmethod
    def get_format_id(cls) -> str:
        return 'netcdf'

    def get_open_data_params_schema(self, data_id: str = None) \
            -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(
            NETCDF_OPEN_DATA_PARAMS_SCHEMA
        )

    def open_data(self,
                  data_id: str,
                  **open_params) -> xr.Dataset:
        assert_instance(data_id, str, name='data_id')
        fs, root, open_params = self.load_fs(open_params)

        # This doesn't yet work as expected with fsspec and netcdf:
        # engine = open_params.pop('engine', 'scipy')
        # with fs.open(data_id, 'rb') as file:
        #     return xr.open_dataset(file, engine=engine, **open_params)

        is_local = is_local_fs(fs)
        if is_local:
            file_path = data_id
        else:
            _, file_path = new_temp_file(suffix='.nc')
            fs.get_file(data_id, file_path)
        engine = open_params.pop('engine', 'netcdf4')
        return xr.open_dataset(file_path, engine=engine, **open_params)

    def get_write_data_params_schema(self) -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(
            NETCDF_WRITE_DATA_PARAMS_SCHEMA
        )

    def write_data(self,
                   data: xr.Dataset,
                   data_id: str,
                   replace=False,
                   **write_params) -> str:
        assert_instance(data, xr.Dataset, name='data')
        assert_instance(data_id, str, name='data_id')
        fs, root, write_params = self.load_fs(write_params)
        if not replace and fs.exists(data_id):
            raise DataStoreError(f'Data resource {data_id} already exists')

        # This doesn't yet work as expected with fsspec and netcdf:
        # engine = write_params.pop('engine', 'scipy')
        # with fs.open(data_id, 'wb') as file:
        #     data.to_netcdf(file, engine=engine, **write_params)

        is_local = is_local_fs(fs)
        if is_local:
            file_path = data_id
        else:
            _, file_path = new_temp_file(suffix='.nc')
        engine = write_params.pop('engine', 'netcdf4')
        data.to_netcdf(file_path, engine=engine, **write_params)
        if not is_local:
            fs.put_file(file_path, data_id)
        return data_id


GEOTIFF_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        tile_size=JsonArraySchema(
            items=(
                JsonNumberSchema(minimum=256,
                                 default=512),
                JsonNumberSchema(minimum=256,
                                 default=512)
            ),
            default=[512, 512]
        ),
        overview_level=JsonIntegerSchema(
            default=None,
            nullable=True,
            description="GeoTIFF overview level. 0 is the first overview."
        )
    ),
    additional_properties=False,
)


class DatasetGeoTiffFsDataAccessor(DatasetFsDataAccessor):
    """
    Opener/writer extension name: "dataset:geotiff:<protocol>"
    """

    @classmethod
    def get_format_id(cls) -> str:
        return 'geotiff'

    def get_open_data_params_schema(self,
                                    data_id: str = None) -> JsonObjectSchema:
        return GEOTIFF_OPEN_DATA_PARAMS_SCHEMA

    def open_data(self,
                  data_id: str,
                  **open_params) -> xr.Dataset:
        assert_instance(data_id, str, name='data_id')
        fs, root, open_params = self.load_fs(open_params)

        if isinstance(fs.protocol, str):
            protocol = fs.protocol
        else:
            protocol = fs.protocol[0]
        if root is not None:
            file_path = protocol + "://" + root + "/" + data_id
        else:
            file_path = protocol + "://" + data_id
        tile_size = open_params.get("tile_size", (512, 512))
        overview_level = open_params.get("overview_level", None)
        return self.open_dataset(fs, file_path, tile_size,
                                 overview_level=overview_level)

    @classmethod
    def create_env_session(cls, fs):
        if isinstance(fs, s3fs.S3FileSystem):
            aws_unsigned = bool(fs.anon)
            aws_session = AWSSession(
                aws_unsigned=aws_unsigned,
                aws_secret_access_key=fs.secret,
                aws_access_key_id=fs.key,
                aws_session_token=fs.token,
                region_name=fs.client_kwargs.get(
                    'region_name',
                    'eu-central-1'
                ),
            )
            return rasterio.env.Env(session=aws_session,
                                    aws_no_sign_request=aws_unsigned
                                    )
        return rasterio.env.NullContextManager()

    @classmethod
    def open_dataset_with_rioxarray(cls, file_path, overview_level,
                                    tile_size) -> rioxarray.raster_array:
        return rioxarray.open_rasterio(
            file_path,
            overview_level=overview_level,
            chunks=dict(zip(('x', 'y'), tile_size))
        )

    @classmethod
    def open_dataset(cls,
                     fs,
                     file_path: str,
                     tile_size: Tuple[int, int],
                     overview_level: Optional[int] = None) -> xr.Dataset:
        """
        A method to open the cog/geotiff dataset using rioxarray,
        returns xarray.Dataset
        @param fs: abstract file system
        @type fs: fsspec.AbstractFileSystem object.
        @param file_path: path of the file
        @type file_path: str
        @param overview_level: the overview level of GeoTIFF, 0 is the first
               overview and None means full resolution.
        @type overview_level: int
        @param tile_size: tile size as tuple.
        @type tile_size: tuple
        """

        if isinstance(fs, fsspec.AbstractFileSystem):
            with cls.create_env_session(fs):
                array = cls.open_dataset_with_rioxarray(file_path,
                                                        overview_level,
                                                        tile_size)
        else:
            assert_true(fs is None, message="invalid type for fs")
        arrays = {}
        if array.ndim == 3:
            for i in range(array.shape[0]):
                name = f'{array.name or "band"}_{i + 1}'
                dims = array.dims[-2:]
                coords = {n: v
                          for n, v in array.coords.items()
                          if n in dims or n == 'spatial_ref'}
                band_data = array.data[i, :, :]
                arrays[name] = xr.DataArray(band_data,
                                            coords=coords,
                                            dims=dims,
                                            attrs=dict(**array.attrs))
        elif array.ndim == 2:
            name = f'{array.name or "band"}'
            arrays[name] = array
        else:
            raise RuntimeError('number of dimensions must be 2 or 3')

        dataset = xr.Dataset(arrays, attrs=dict(source=file_path))
        # For CRS, rioxarray uses variable "spatial_ref" by default
        if 'spatial_ref' in array.coords:
            for data_var in dataset.data_vars.values():
                data_var.attrs['grid_mapping'] = 'spatial_ref'

        # rioxarray may return non-JSON-serializable metadata
        # attribute values.
        # We have seen _FillValue of type np.uint8
        cls._sanitize_dataset_attrs(dataset)

        return dataset

    def get_write_data_params_schema(self) -> JsonObjectSchema:
        raise NotImplementedError("Writing of GeoTIFF not yet supported")

    def write_data(self,
                   data: xr.Dataset,
                   data_id: str,
                   replace=False,
                   **write_params) -> str:
        raise NotImplementedError("Writing of GeoTIFF not yet supported")

    @classmethod
    def _sanitize_dataset_attrs(cls, dataset):
        dataset.attrs.update(to_json_value(dataset.attrs))
        for var in dataset.variables.values():
            var.attrs.update(to_json_value(var.attrs))



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from ..accessor import COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES
from ..accessor import FsAccessor


class FileFsAccessor(FsAccessor):

    @classmethod
    def get_protocol(cls) -> str:
        return 'file'

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                auto_mkdirs=JsonBooleanSchema(
                    description='Whether, when opening a file, the directory'
                                ' containing it should be created (if it'
                                ' doesn\'t already exist).'),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES
            ),
            additional_properties=True,
        )


class MemoryFsAccessor(FsAccessor):

    @classmethod
    def get_protocol(cls) -> str:
        return 'memory'


class S3FsAccessor(FsAccessor):

    @classmethod
    def get_protocol(cls) -> str:
        return 's3'

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        # We may use here AWS S3 defaults as described in
        #   https://boto3.amazonaws.com/v1/documentation/api/
        #   latest/guide/configuration.html
        return JsonObjectSchema(
            properties=dict(
                anon=JsonBooleanSchema(
                    title='Whether to anonymously connect to AWS S3.'
                ),
                key=JsonStringSchema(
                    min_length=1,
                    title='AWS access key identifier.',
                    description='Can also be set in profile section'
                                ' of ~/.aws/config, or by environment'
                                ' variable AWS_ACCESS_KEY_ID.'
                ),
                secret=JsonStringSchema(
                    min_length=1,
                    title='AWS secret access key.',
                    description='Can also be set in profile section'
                                ' of ~/.aws/config, or by environment'
                                ' variable AWS_SECRET_ACCESS_KEY.'
                ),
                token=JsonStringSchema(
                    min_length=1,
                    title='Session token.',
                    description='Can also be set in profile section'
                                ' of ~/.aws/config, or by environment'
                                ' variable AWS_SESSION_TOKEN.'
                ),
                use_ssl=JsonBooleanSchema(
                    description='Whether to use SSL in connections to S3;'
                                ' may be faster without, but insecure.',
                    default=True,
                ),
                requester_pays=JsonBooleanSchema(
                    description='If "RequesterPays" buckets are supported.',
                    default=False,
                ),
                s3_additional_kwargs=JsonObjectSchema(
                    description='parameters that are used when calling'
                                ' S3 API methods. Typically used for'
                                ' things like "ServerSideEncryption".',
                    additional_properties=True,
                ),
                client_kwargs=JsonObjectSchema(
                    description='Parameters for the botocore client.',
                    properties=dict(
                        endpoint_url=JsonStringSchema(
                            min_length=1,
                            format='uri',
                            title='Alternative endpoint URL.'
                        ),
                        # bucket_name=JsonStringSchema(
                        #     min_length=1,
                        #     title='Name of the bucket'
                        # ),
                        profile_name=JsonStringSchema(
                            min_length=1,
                            title='Name of the AWS configuration profile',
                            description='Section name with within'
                                        ' ~/.aws/config file,'
                                        ' which provides AWS configurations'
                                        ' and credentials.'
                        ),
                        region_name=JsonStringSchema(
                            min_length=1,
                            title='AWS storage region name'
                        ),
                    ),
                    additional_properties=True,
                ),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            ),
            additional_properties=True,
        )


class AzureFsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return 'abfs'

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                anon=JsonBooleanSchema(
                    title='Whether to anonymously connect to'
                          ' Azure Blob Storage.'
                ),
                account_name=JsonStringSchema(
                    min_length=1,
                    title='Azure storage account name.',
                    description='Must be used with the account key parameter.'
                                ' This is not required when using a'
                                ' connection string.'

                ),
                account_key=JsonStringSchema(
                    min_length=1,
                    title='Azure storage account key.',
                    description='Must be used with the account'
                                ' name parameter.'
                                ' This is not required when using a'
                                ' connection string'
                ),
                connection_string=JsonStringSchema(
                    min_length=1,
                    title='Connection string for Azure blob storage.',
                    description='Use this parameter inplace of both'
                                ' account name and key'
                                ' because they are both contained'
                                ' in the string.'
                ),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            ),
            additional_properties=True,
        )


class FtpFsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return 'ftp'

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                host=JsonStringSchema(
                    title='FTP host',
                    description='The remote server name/ip to connect to',
                    min_length=1
                ),
                port=JsonIntegerSchema(
                    minimum=0,
                    maximum=65535,
                    default=21,
                    title='FTP port',
                    description='Port to connect with'
                ),
                username=JsonStringSchema(
                    min_length=1,
                    title='User name',
                    description='If authenticating, the user\'s identifier'
                ),
                password=JsonStringSchema(
                    min_length=1,
                    title='User password',
                    description='User\'s password on the server, if using',
                ),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            ),
            additional_properties=True,
        )




# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


from xcube.server.api import Context
from xcube.webapi.common.context import ResourcesContext

from ...datasets.context import DatasetsContext


class StacContext(ResourcesContext):

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self._datasets_ctx = server_ctx.get_api_ctx('datasets',
                                                    cls=DatasetsContext)

        # check: determine what else to include in this list.
        #  Listing all 11396 CRSs currently known to pyproj seems
        #  impractical. Make it a configuration option, perhaps?
        self._available_crss = ['OGC:CRS84', 'EPSG:4326']

    @property
    def datasets_ctx(self) -> DatasetsContext:
        return self._datasets_ctx

    @property
    def available_crss(self) -> list[str]:
        return self._available_crss



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from abc import ABCMeta, abstractmethod
from typing import Any, Collection, Dict, Mapping, Optional, Tuple, Union

import numpy as np
import xarray as xr

from xcube.constants import CRS_WKT_EPSG_4326
from xcube.constants import EXTENSION_POINT_INPUT_PROCESSORS
from xcube.core.gridmapping import GridMapping
from xcube.core.resampling import rectify_dataset
from xcube.core.timecoord import get_time_range_from_attrs
from xcube.core.timecoord import get_time_range_from_data
from xcube.core.timecoord import to_time_in_days_since_1970
from xcube.util.plugin import ExtensionComponent
from xcube.util.plugin import get_extension_registry


class ReprojectionInfo:
    """
    Characterize input datasets so we can reproject.

    :param xy_names: Names of variables providing the spatial x- and y-coordinates,
           e.g. ('longitude', 'latitude')
    :param xy_tp_names: Optional names of tie-point variables providing the spatial y- and y-coordinates,
           e.g. ('TP_longitude', 'TP_latitude')
    :param xy_crs: Optional spatial reference system, e.g. 'EPSG:4326' or WKT or proj4 mapping
    :param xy_gcp_step: Optional step size for collecting ground control points from spatial
           coordinate arrays denoted by *xy_names*.
    :param xy_tp_gcp_step: Optional step size for collecting ground control points from spatial
           coordinate arrays denoted by *xy_tp_names*.
    """

    def __init__(self,
                 xy_names: Tuple[str, str] = None,
                 xy_tp_names: Tuple[str, str] = None,
                 xy_crs: Any = None,
                 xy_gcp_step: Union[int, Tuple[int, int]] = None,
                 xy_tp_gcp_step: Union[int, Tuple[int, int]] = None):
        self._xy_names = self._assert_name_pair('xy_names', xy_names)
        self._xy_tp_names = self._assert_name_pair('xy_tp_names', xy_tp_names)
        self._xy_crs = xy_crs
        self._xy_gcp_step = self._assert_step_pair('xy_gcp_step', xy_gcp_step)
        self._xy_tp_gcp_step = self._assert_step_pair('xy_tp_gcp_step', xy_tp_gcp_step)

    def derive(self,
               xy_names: Tuple[str, str] = None,
               xy_tp_names: Tuple[str, str] = None,
               xy_crs: Any = None,
               xy_gcp_step: Union[int, Tuple[int, int]] = None,
               xy_tp_gcp_step: Union[int, Tuple[int, int]] = None):
        return ReprojectionInfo(self.xy_names if xy_names is None else xy_names,
                                xy_tp_names=self.xy_tp_names if xy_tp_names is None else xy_tp_names,
                                xy_crs=self.xy_crs if xy_crs is None else xy_crs,
                                xy_gcp_step=self.xy_gcp_step if xy_gcp_step is None else xy_gcp_step,
                                xy_tp_gcp_step=self.xy_tp_gcp_step if xy_tp_gcp_step is None else xy_tp_gcp_step)

    @property
    def xy_names(self) -> Optional[Tuple[str, str]]:
        return self._xy_names

    @property
    def xy_tp_names(self) -> Optional[Tuple[str, str]]:
        return self._xy_tp_names

    @property
    def xy_crs(self) -> Any:
        return self._xy_crs

    @property
    def xy_gcp_step(self) -> Optional[int]:
        return self._xy_gcp_step

    @property
    def xy_tp_gcp_step(self) -> Optional[int]:
        return self._xy_tp_gcp_step

    def _assert_name_pair(self, keyword: str, value):
        if value is not None:
            v1, v2 = value
            self._assert_name(keyword, v1)
            self._assert_name(keyword, v2)
            return v1, v2
        return value

    def _assert_step_pair(self, keyword: str, value):
        if value is not None:
            if isinstance(value, int):
                v1, v2 = value, value
            else:
                v1, v2 = value
            self._assert_step(keyword, v1)
            self._assert_step(keyword, v2)
            return v1, v2
        return value

    def _assert_name(self, keyword: str, value):
        if value is None:
            raise ValueError(f'invalid {keyword}, missing name')
        if not isinstance(value, str) or not value:
            raise ValueError(f'invalid {keyword}, name must be a non-empty string')

    def _assert_step(self, keyword: str, value):
        if value is None:
            raise ValueError(f'invalid {keyword}, missing name')
        if not isinstance(value, int) or value <= 0:
            raise ValueError(f'invalid {keyword}, step must be an integer number')


class InputProcessor(ExtensionComponent, metaclass=ABCMeta):
    """
    Read and process inputs for the gen tool.

    An InputProcessor can be configured by the following parameters:

    * ``input_reader``: The input format identifier. Required, no default.

    :param name: A unique input processor identifier.
    """

    def __init__(self, name: str, **parameters):
        super().__init__(EXTENSION_POINT_INPUT_PROCESSORS, name)
        self._parameters = {**self.default_parameters, **parameters}
        if 'input_reader' not in self._parameters:
            raise ValueError('missing input_reader in input processor parameters')

    @property
    def description(self) -> str:
        """
        :return: A description for this input processor
        """
        return self.get_metadata_attr('description', '')

    @property
    def default_parameters(self) -> Dict[str, Any]:
        return {}

    @property
    def parameters(self) -> Mapping[str, Any]:
        return self._parameters

    @property
    def input_reader(self) -> str:
        return self.parameters['input_reader']

    @property
    def input_reader_params(self) -> dict:
        """
        :return: The input reader parameters for this input processor.
        """
        return self.parameters.get('input_reader_params', {})

    @abstractmethod
    def get_time_range(self, dataset: xr.Dataset) -> Optional[Tuple[float, float]]:
        """
        Return a tuple of two floats representing start/stop time (which may be same) in days since 1970.
        :param dataset: The dataset.
        :return: The time-range tuple of the dataset or None.
        """
        raise NotImplementedError()

    def get_extra_vars(self, dataset: xr.Dataset) -> Optional[Collection[str]]:
        """
        Get a set of names of variables that are required as input for the pre-processing and processing
        steps and should therefore not be dropped.
        However, the processing or post-processing steps may later remove them.

        Returns ``None`` by default.

        :param dataset: The dataset.
        :return: Collection of names of variables to be prevented from being dropping.
        """
        return None

    def pre_process(self, dataset: xr.Dataset) -> xr.Dataset:
        """
        Do any pre-processing before reprojection.
        All variables in the output dataset must be 2D arrays with dimensions "lat" and "lon", in this order.
        For example, perform dataset validation, masking, and/or filtering using provided configuration parameters.

        The default implementation returns the unchanged *dataset*.

        :param dataset: The dataset.
        :return: The pre-processed dataset or the original one, if no pre-processing is required.
        """
        return dataset

    # Note, arg "geo_coding" should be called "source_gm",
    # arg "output_geom" should be called "target_gm",
    # but this package will be deprecated anyway.
    # See #540

    @abstractmethod
    def process(self,
                dataset: xr.Dataset,
                geo_coding: GridMapping,
                output_geom: GridMapping,
                output_resampling: str,
                include_non_spatial_vars=False) -> xr.Dataset:
        """
        Perform spatial transformation into the cube's WGS84 SRS such that all variables in the output dataset
        * must be 2D arrays with dimensions "lat" and "lon", in this order, and
        * must have shape (*dst_size[-1]*, *dst_size[-2]*), and
        * must have *dst_region* as their bounding box in geographic coordinates.

        :param dataset: The input dataset.
        :param geo_coding: The input's geo-coding.
        :param output_geom: The output's spatial image geometry.
        :param output_resampling: The output's spatial resampling method.
        :param include_non_spatial_vars: Whether to include non-spatial variables in the output.
        :return: The transformed output dataset or the original one, if no transformation is required.
        """
        raise NotImplementedError()

    def post_process(self, dataset: xr.Dataset) -> xr.Dataset:
        """
        Do any post-processing transformation. The input is a 3D array with dimensions ("time", "lat", "lon").
        Post-processing may, for example, generate new "wavelength" dimension for variables whose name follow
        a certain pattern.

        The default implementation returns the unchanged *dataset*.

        :param dataset: The dataset.
        :return: The post-processed dataset or the original one, if no post-processing is required.
        """
        return dataset


class XYInputProcessor(InputProcessor, metaclass=ABCMeta):
    """
    Read and process inputs for the gen tool.

    An XYInputProcessor can be configured by the following parameters:

    * ``input_reader``: The input format identifier.
        Required, no default.
    * ``xy_names``: A tuple of names of the variable providing x,y geo-locations.
        Optional, looked up automatically if not given e.g. ``("lon", "lat")``.
    * ``xy_tp_names``: A tuple of names of the variable providing x,y tie-point geo-locations.
        Optional, no default.
    * ``xy_crs``: A WKT string that identifies the x,y coordinate reference system (CRS).
        Optional, no default.
    * ``xy_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    * ``xy_tp_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y tie-point coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    """

    @property
    def default_parameters(self) -> Dict[str, Any]:
        default_parameters = super().default_parameters
        default_parameters.update(xy_names=('lon', 'lat'))
        return default_parameters

    def get_reprojection_info(self, dataset: xr.Dataset) -> ReprojectionInfo:
        """
        Information about special fields in input datasets used for reprojection.
        :param dataset: The dataset.
        :return: The reprojection information of the dataset or None.
        """
        parameters = self.parameters
        return ReprojectionInfo(xy_names=parameters.get('xy_names', ('lon', 'lat')),
                                xy_tp_names=parameters.get('xy_tp_names'),
                                xy_crs=parameters.get('xy_crs'),
                                xy_gcp_step=parameters.get('xy_gcp_step'),
                                xy_tp_gcp_step=parameters.get('xy_tp_gcp_step'))

    def get_extra_vars(self, dataset: xr.Dataset) -> Optional[Collection[str]]:
        """
        Return the names of variables containing spatial coordinates.
        They should not be removed, as they are required for the reprojection.
        """
        reprojection_info = self.get_reprojection_info(dataset)
        if reprojection_info is None:
            return dataset
        extra_vars = set()
        if reprojection_info.xy_names:
            extra_vars.update(set(reprojection_info.xy_names))
        if reprojection_info.xy_tp_names:
            extra_vars.update(set(reprojection_info.xy_tp_names))
        return extra_vars

    def process(self,
                dataset: xr.Dataset,
                geo_coding: GridMapping,
                output_geom: GridMapping,
                output_resampling: str,
                include_non_spatial_vars=False) -> xr.Dataset:
        """
        Perform reprojection using tie-points / ground control points.
        """
        reprojection_info = self.get_reprojection_info(dataset)

        warn_prefix = 'unsupported argument in np-GCP rectification mode'
        if reprojection_info.xy_crs is not None:
            warnings.warn(f'{warn_prefix}: ignoring '
                          f'reprojection_info.xy_crs = {reprojection_info.xy_crs!r}')
        if reprojection_info.xy_tp_names is not None:
            warnings.warn(f'{warn_prefix}: ignoring '
                          f'reprojection_info.xy_tp_names = {reprojection_info.xy_tp_names!r}')
        if reprojection_info.xy_gcp_step is not None:
            warnings.warn(f'{warn_prefix}: ignoring '
                          f'reprojection_info.xy_gcp_step = {reprojection_info.xy_gcp_step!r}')
        if reprojection_info.xy_tp_gcp_step is not None:
            warnings.warn(f'{warn_prefix}: ignoring '
                          f'reprojection_info.xy_tp_gcp_step = {reprojection_info.xy_tp_gcp_step!r}')
        if output_resampling != 'Nearest':
            warnings.warn(f'{warn_prefix}: ignoring '
                          f'dst_resampling = {output_resampling!r}')
        if include_non_spatial_vars:
            warnings.warn(f'{warn_prefix}: ignoring '
                          f'include_non_spatial_vars = {include_non_spatial_vars!r}')

        geo_coding = geo_coding.derive(xy_var_names=(reprojection_info.xy_names[0],
                                                     reprojection_info.xy_names[1]))

        dataset = rectify_dataset(dataset,
                                  compute_subset=False,
                                  source_gm=geo_coding,
                                  target_gm=output_geom)
        if output_geom.is_tiled:
            # The following condition may become true,
            # if we have used rectified_dataset(input, ..., is_y_reverse=True)
            # In this case y-chunksizes will also be reversed. So that the first chunk is smaller than any other.
            # Zarr will reject such datasets, when written.
            if dataset.chunks.get('lat')[0] < dataset.chunks.get('lat')[-1]:
                dataset = dataset.chunk({'lat': output_geom.tile_height,
                                         'lon': output_geom.tile_width})
        if dataset is not None \
                and geo_coding.crs.is_geographic \
                and geo_coding.xy_var_names != ('lon', 'lat'):
            dataset = dataset.rename({geo_coding.xy_var_names[0]: 'lon',
                                      geo_coding.xy_var_names[1]: 'lat'})

        return dataset


class DefaultInputProcessor(XYInputProcessor):
    """
    Default input processor that expects input datasets to have the xcube standard format:

    * Have dimensions ``lat``, ``lon``, optionally ``time`` of length 1;
    * have coordinate variables ``lat[lat]``, ``lon[lat]``, ``time[time]`` (opt.), ``time_bnds[time, 2]`` (opt.);
    * have coordinate variables ``lat[lat]``, ``lon[lat]`` as decimal degrees on WGS84 ellipsoid,
      both linearly increasing with same constant delta;
    * have coordinate variable ``time[time]`` representing a date+time values with defined CF "units" attribute;
    * have any data variables of form ``<var>[time, lat, lon]``;
    * have global attribute pairs (``time_coverage_start``, ``time_coverage_end``), or (``start_time``, ``stop_time``)
      if ``time`` coordinate is missing.

    The default input processor can be configured by the following parameters:

    * ``input_reader``: The input format identifier.
        Required, defaults to ``"netcdf4"``.
    * ``xy_names``: A tuple of names of the variable providing x,y geo-locations.
        Optional, defaults to ``("lon", "lat")``.
    * ``xy_tp_names``: A tuple of names of the variable providing x,y tie-point geo-locations.
        Optional, no default.
    * ``xy_crs``: A WKT string that identifies the x,y coordinate reference system (CRS).
        Optional, defaults to WKT for EPSG:4326 (see ``xcube.constants.CRS_WKT_EPSG_4326`` constant).
    * ``xy_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    * ``xy_tp_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y tie-point coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    """

    def __init__(self, **parameters):
        super().__init__('default', **parameters)

    @property
    def default_parameters(self) -> Dict[str, Any]:
        default_parameters = super().default_parameters
        default_parameters.update(input_reader='netcdf4',
                                  xy_names=('lon', 'lat'),
                                  xy_crs=CRS_WKT_EPSG_4326)
        return default_parameters

    def pre_process(self, dataset: xr.Dataset) -> xr.Dataset:
        self._validate(dataset)

        if "time" in dataset.dims:
            # Remove time dimension of length 1.
            dataset = dataset.squeeze("time")

        return _normalize_lon_360(dataset)

    def get_time_range(self, dataset: xr.Dataset) -> Tuple[float, float]:
        time_coverage_start, time_coverage_end = get_time_range_from_data(dataset)
        if time_coverage_start is not None:
            time_coverage_start = str(time_coverage_start)
        if time_coverage_end is not None:
            time_coverage_end = str(time_coverage_end)
        if time_coverage_start is None or time_coverage_end is None:
            time_coverage_start, time_coverage_end = get_time_range_from_attrs(dataset)
        if time_coverage_start is None:
            raise ValueError("invalid input: missing time coverage information in dataset")
        if time_coverage_end is None:
            time_coverage_end = time_coverage_start
        return to_time_in_days_since_1970(time_coverage_start), \
               to_time_in_days_since_1970(time_coverage_end)

    def _validate(self, dataset):
        self._check_coordinate_var(dataset, "lon", min_length=2)
        self._check_coordinate_var(dataset, "lat", min_length=2)
        if "time" in dataset.dims:
            self._check_coordinate_var(dataset, "time", max_length=1)
            required_dims = ("time", "lat", "lon")
        else:
            required_dims = ("lat", "lon")
        count = 0
        for var_name in dataset.data_vars:
            var = dataset.data_vars[var_name]
            if var.dims == required_dims:
                count += 1
        if count == 0:
            raise ValueError(f"dataset has no variables with required dimensions {required_dims!r}")

    # noinspection PyMethodMayBeStatic
    def _check_coordinate_var(self, dataset: xr.Dataset, coord_var_name: str,
                              min_length: int = None, max_length: int = None):
        if coord_var_name not in dataset.coords:
            raise ValueError(f'missing coordinate variable "{coord_var_name}"')
        coord_var = dataset.coords[coord_var_name]
        if len(coord_var.shape) != 1:
            raise ValueError('coordinate variable "lon" must be 1D')
        coord_var_bnds_name = coord_var.attrs.get("bounds", coord_var_name + "_bnds")
        if coord_var_bnds_name in dataset:
            coord_bnds_var = dataset[coord_var_bnds_name]
            expected_shape = (len(coord_var), 2)
            if coord_bnds_var.shape != expected_shape:
                raise ValueError(f'coordinate bounds variable "{coord_bnds_var}" must have shape {expected_shape!r}')
        else:
            if min_length is not None and len(coord_var) < min_length:
                raise ValueError(f'coordinate variable "{coord_var_name}" must have at least {min_length} value(s)')
            if max_length is not None and len(coord_var) > max_length:
                raise ValueError(f'coordinate variable "{coord_var_name}" must have no more than {max_length} value(s)')


def _normalize_lon_360(dataset: xr.Dataset) -> xr.Dataset:
    """
    Fix the longitude of the given dataset ``dataset`` so that it ranges from -180 to +180 degrees.

    :param dataset: The dataset whose longitudes may be given in the range 0 to 360.
    :return: The fixed dataset or the original dataset.
    """

    if 'lon' not in dataset.coords:
        return dataset

    lon_var = dataset.coords['lon']

    if len(lon_var.shape) != 1:
        return dataset

    lon_size = lon_var.shape[0]
    if lon_size < 2:
        return dataset

    lon_size_05 = lon_size // 2
    lon_values = lon_var.values
    if not np.any(lon_values[lon_size_05:] > 180.):
        return dataset

    # roll_coords will be set to False by default in the future
    dataset = dataset.roll(lon=lon_size_05, roll_coords=True)
    dataset = dataset.assign_coords(lon=(((dataset.lon + 180) % 360) - 180))

    return dataset


def find_input_processor_class(name: str):
    extension = get_extension_registry().get_extension(EXTENSION_POINT_INPUT_PROCESSORS, name)
    if not extension:
        return None
    return extension.component



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import os.path
from typing import Callable

import click

from xcube.cli.common import (cli_option_quiet,
                              cli_option_verbosity,
                              configure_cli_output)
from xcube.constants import (FORMAT_NAME_ZARR,
                             LOG,
                             LOG_LEVEL_DETAIL)

Monitor = Callable[[str, int], None]


# noinspection PyShadowingBuiltins
@click.command(name='prune')
@click.argument('dataset_path', metavar='DATASET')
@cli_option_quiet
@cli_option_verbosity
@click.option('--dry-run', is_flag=True,
              help='Just read and process input, '
                   'but don\'t produce any output.')
def prune(dataset_path: str,
          quiet: bool,
          verbosity: int,
          dry_run: bool):
    """
    Delete empty chunks.
    Deletes all data files associated with empty
    (NaN-only) chunks in given DATASET,
    which must have Zarr format.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    def monitor(msg: str, monitor_verbosity: int = 1):
        if monitor_verbosity == 0:
            LOG.error(msg)
        elif monitor_verbosity == 1:
            LOG.info(msg)
        elif monitor_verbosity == 2:
            LOG.log(LOG_LEVEL_DETAIL, msg)
        elif monitor_verbosity == 3:
            LOG.debug(msg)

    _prune(input_path=dataset_path, dry_run=dry_run, monitor=monitor)
    return 0


def _prune(input_path: str,
           dry_run: bool,
           monitor: Monitor):
    from xcube.core.chunk import get_empty_dataset_chunks
    from xcube.core.dsio import guess_dataset_format
    from xcube.core.dsio import open_dataset

    input_format = guess_dataset_format(input_path)
    if input_format != FORMAT_NAME_ZARR:
        raise click.ClickException("input must be a dataset in Zarr format")

    num_deleted_total = 0

    monitor(f'Opening dataset from {input_path!r}...', 1)
    with open_dataset(input_path) as dataset:
        monitor('Identifying empty chunks...', 1)
        for var_name, chunk_indices in get_empty_dataset_chunks(dataset):
            num_empty_chunks = 0
            num_deleted = 0
            for chunk_index in chunk_indices:
                num_empty_chunks += 1
                if num_empty_chunks == 1:
                    monitor(f'Found empty chunks in variable {var_name!r}, '
                            f'deleting block files...', 2)

                ok = _delete_block_file(input_path, var_name, chunk_index,
                                        dry_run, monitor)
                if ok:
                    num_deleted += 1
            if num_deleted > 0:
                monitor(f'Deleted {num_deleted} block file(s) '
                        f'for variable {var_name!r}.', 2)
            elif num_empty_chunks > 0:
                monitor(f'No block files for variable {var_name!r} '
                        f'could be deleted.', 2)
            num_deleted_total += num_deleted

    monitor(f'Done, {num_deleted_total} block file(s) deleted total.', 1)


def _delete_block_file(input_path: str,
                       var_name: str,
                       chunk_index,
                       dry_run: bool,
                       monitor: Monitor) -> bool:
    block_path = None
    block_path_1 = os.path.join(input_path,
                                var_name,
                                '.'.join(map(str, chunk_index)))
    block_path_2 = block_path_1

    if os.path.isfile(block_path_1):
        block_path = block_path_1
    else:
        block_path_2 = os.path.join(input_path,
                                    var_name,
                                    *map(str, chunk_index))
        if os.path.isfile(block_path_2):
            block_path = block_path_2

    if block_path:
        monitor(f'Deleting chunk file {block_path}', 3)
        if dry_run:
            return True
        try:
            os.remove(block_path)
            return True
        except OSError as e:
            monitor(f'Failed to delete '
                    f'block file {block_path}: {e}', 0)
    else:
        monitor(f'Could find neither '
                f'block file {block_path_1} nor {block_path_2}', 0)
    return False



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.schema import rechunk_cube
from .transformer import CubeTransformer
from .transformer import TransformedCube
from ..config import CubeConfig


class CubeRechunker(CubeTransformer):
    """Force cube to have chunks compatible with Zarr."""

    def transform_cube(self,
                       cube: xr.Dataset,
                       gm: GridMapping,
                       cube_config: CubeConfig) -> TransformedCube:
        cube, gm = rechunk_cube(cube, gm,
                                chunks=cube_config.chunks,
                                tile_size=cube_config.tile_size)
        return cube, gm, cube_config



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import importlib
import inspect
import os.path
import sys
import warnings
from typing import Any, Union, Dict, Callable, Optional, Tuple

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from xcube.util.temp import new_temp_dir
from .constants import DEFAULT_CALLABLE_NAME
from .constants import DEFAULT_MODULE_NAME
from .constants import TEMP_FILE_PREFIX
from .fileset import FileSet
from ...constants import LOG


class CodeConfig(JsonObject):
    """
    Code configuration object.

    Instances should always be created using one of the factory methods:

    * :meth:from_callable
    * :meth:from_code
    * :meth:from_file_set

    :param _callable: Optional function or class.
        Cannot be given if *inline_code* or *file_set* are given.
    :param callable_ref: Optional reference to the callable
        in the *file_set*. Must have form "<module-name>:<callable-name>".
    :param inline_code: Optional inline Python code string.
        Cannot be given if *_callable* or *file_set* are given.
    :param file_set: A file set that contains Python
        modules or packages.
        Must be of type :class:FileSet.
        Cannot be given if *_callable* or *inline_code* are given.
    :param install_required: Whether *file_set* contains
        Python modules or packages that must be installed.
        Can only be True if *file_set* is given.
    :param callable_params: Optional parameters that are supposed to
        be passed as keyword-arguments to the callable.
        Stored as part of the configuration, but this class
        does not apply them.
        Must be a dictionary, if given.
    """

    def __init__(self,
                 _callable: Callable = None,
                 callable_ref: str = None,
                 callable_params: Dict[str, Any] = None,
                 inline_code: str = None,
                 file_set: FileSet = None,
                 install_required: bool = None):
        if callable_ref is not None:
            # Will validate callable_ref
            _normalize_callable_ref(callable_ref)
        if _callable is not None:
            assert_true(callable(_callable),
                        '_callable must be a function or class')
        if callable_params is not None:
            assert_instance(callable_params, dict, 'callable_params')
        if inline_code is not None:
            assert_instance(inline_code, str, 'inline_code')
        if file_set is not None:
            assert_instance(file_set, FileSet, 'file_set')
        assert_true(sum((1 if v is not None else 0) for v in
                        (_callable, inline_code, file_set)) == 1,
                    '_callable, inline_code, file_set '
                    'are mutually exclusive and one must be given')
        assert_true(not install_required or file_set is not None,
                    'install_required can only be used if file_set is given')
        self._callable = _callable
        self.callable_ref = callable_ref
        self.callable_params = callable_params
        self.inline_code = inline_code
        self.file_set = file_set
        self.install_required = install_required

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        """Get the JSON schema for CodeConfig objects."""
        return JsonObjectSchema(
            properties=dict(
                callable_ref=JsonStringSchema(min_length=1),
                callable_params=JsonObjectSchema(additional_properties=True),
                inline_code=JsonStringSchema(min_length=1),
                file_set=FileSet.get_schema(),
                install_required=JsonBooleanSchema(),
            ),
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_code(cls,
                  *code: Union[str, Callable],
                  callable_name: str = None,
                  module_name: str = None,
                  callable_params: Dict[str, Any] = None) -> 'CodeConfig':
        """
        Create a code configuration from the given *code* which may be
        given as one or more plain text strings or callables.

        This will create a configuration that uses an inline
        ``code_string`` which contains the source code.

        :param code: The code.
        :param callable_name: The callable name.
            If not given, will be inferred from first callable.
            Otherwise, it defaults to "process_dataset".
        :param module_name: The module name. If not given,
            defaults to "user_code".
        :param callable_params: Optional parameters that are supposed to
            be passed as keyword-arguments to the callable.
            Stored as part of the configuration, but this class
            does not apply them. Must be a dictionary, if given.
        :return: A new code configuration.
        """
        assert_given(code, 'code')
        inline_code, callable_ref = _normalize_inline_code(
            code,
            module_name=module_name,
            callable_name=callable_name
        )
        return CodeConfig(callable_ref=callable_ref,
                          inline_code=inline_code,
                          callable_params=callable_params)

    @classmethod
    def from_callable(
            cls,
            _callable: Callable,
            callable_params: Optional[Dict[str, Any]] = None
    ) -> 'CodeConfig':
        """
        Create a code configuration from the callable *_callable*.

        Note, the resulting code configuration is only
        valid in a local context. It cannot be JSON-serialized.

        To pass such configurations to a service, convert it first
        using the :meth:to_service first.

        :param _callable: A function or class
        :param callable_params: Optional parameters that are supposed to
            be passed as keyword-arguments to the callable.
            Stored as part of the configuration, but this class
            does not apply them. Must be a dictionary, if given.
        :return: A new code configuration.
        """
        assert_given(_callable, '_callable')
        return CodeConfig(_callable=_callable,
                          callable_params=callable_params)

    @classmethod
    def from_file_set(cls,
                      file_set: Union[FileSet, str, Any],
                      callable_ref: str,
                      callable_params: Optional[Dict[str, Any]] = None,
                      install_required: Optional[bool] = None) \
            -> 'CodeConfig':
        """
        Create a code configuration from a file set.

        :param file_set: The file set.
            Can be a path or a :class:FileSet instance.
        :param callable_ref: Reference to the callable in the *file_set*,
            must have form "<module-name>:<callable-name>"
        :param install_required: Whether the *file_set* is
            a package that must be installed.
        :param callable_params: Optional parameters that are supposed to
            be passed as keyword-arguments to the callable.
            Stored as part of the configuration, but this class
            does not apply them. Must be a dictionary, if given.
        :return: A new code configuration.
        """
        assert_given(file_set, 'file_set')
        assert_given(callable_ref, 'callable_ref')
        return CodeConfig(callable_ref=callable_ref,
                          file_set=_normalize_file_set(file_set),
                          install_required=install_required,
                          callable_params=callable_params)

    @classmethod
    def from_github_archive(cls,
                            gh_org: str,
                            gh_repo: str,
                            gh_tag: str,
                            gh_release: str,
                            callable_ref: str,
                            callable_params: Optional[Dict[str, Any]] = None,
                            gh_username: Optional[str] = None,
                            gh_token: Optional[str] = None):
        """
        Create a code configuration from a GitHub archive.

        :param gh_org: GitHub organisation name or username
        :param gh_repo:  GitHub repository name
        :param gh_tag: GitHub release tag
        :param gh_release: The name of a GitHub release. It is used to
            form the sub-path into the archive. The sub-path has the form
            "<gh_repo>-<gh_release>".
        :param callable_ref: Reference to the callable in the *file_set*,
            must have form "<module-name>:<callable-name>"
        :param callable_params: Optional parameters that are supposed to
            be passed as keyword-arguments to the callable.
            Stored as part of the configuration, but this class
            does not apply them. Must be a dictionary, if given.
        :param gh_username: Optional GitHub username.
        :param gh_token: Optional GitHub username.
        :return:
        """
        assert_given(gh_org, 'gh_org')
        assert_given(gh_org, 'gh_repo')
        assert_given(gh_org, 'gh_tag')
        assert_given(gh_org, 'gh_release')
        gh_url = f'https://github.com/{gh_org}/{gh_repo}/archive/{gh_tag}.zip'
        gh_sub_path = f'{gh_repo}-{gh_release}'
        gh_params = None
        if gh_token is not None:
            gh_params = dict(token=gh_token, username=gh_username or gh_org)
        elif gh_username is not None:
            assert_given(gh_token, 'gh_token')  # fails always
        return CodeConfig(file_set=FileSet(gh_url,
                                           sub_path=gh_sub_path,
                                           storage_params=gh_params),
                          callable_ref=callable_ref,
                          callable_params=callable_params)

    def for_service(self) -> 'CodeConfig':
        """
        Convert this code configuration so can be used by the
        generator service.

        The returned code configuration defines either
        inline code or a file set that represents a local
        ZIP archive.
        """
        return _for_service(self)

    def for_local(self) -> 'CodeConfig':
        """
        Convert this code configuration so can be used by the
        local generator. This means, the returned configuration
        can be used to load executable code. i.e.
        :meth:get_callable() will return a callable.

        The returned code configuration defines either
        a callable (member _callable) or a file set that represents
        a local directory. This directory may be used without
        package installation (i.e. added to Python's sys.path) or
        used after installation (i.e. by executing
        ```python setup.py install```).
        """
        return _for_local(self)

    def get_callable(self) -> Callable:
        """
        Get the callable specified by this configuration.

        In the common case, this will require importing the callable.

        :return: A callable
        :raise ImportError if the callable can not be imported
        """
        if self._callable is None:
            self.set_callable(self._load_callable())
        return self._callable

    def set_callable(self, func_or_class: Callable):
        """
        Set the callable that is represented by this configuration.
        :param func_or_class: A callable
        """
        assert_true(callable(func_or_class),
                    f'func_or_class must be callable')
        self._callable = func_or_class

    def _load_callable(self) -> Callable:
        """
        Load the callable specified by this configuration.

        :return: A callable
        :raise ImportError if the callable can not be imported
        """
        code_config = self.for_local()
        return _load_callable(code_config.file_set.path,
                              code_config.callable_ref,
                              code_config.install_required)


def _for_service(code_config: CodeConfig) -> 'CodeConfig':
    if code_config.inline_code is not None:
        # If we have inline code,
        # there is no need to do anything else.
        return code_config

    if code_config.file_set is not None \
            and not code_config.file_set.is_local():
        # For service requests, remote file set will
        # simply turn into JSON. for_local() will download
        # them when request is executed.
        # So we are done here.
        return code_config

    # noinspection PyProtectedMember
    if code_config._callable is not None \
            and code_config.inline_code is None \
            and code_config.file_set is None:
        # If we have callable that was not loaded
        # from inline string or files set,
        # turn into a local Python module.
        # noinspection PyProtectedMember
        code_config = _callable_to_module(code_config._callable,
                                          code_config.callable_params)

    if code_config.file_set is not None \
            and code_config.file_set.is_local_dir():
        # If we have a file set that is a local
        # directory, turn it into a local ZIP Archive.
        file_set = code_config.file_set.to_local_zip()
        code_config = code_config.from_file_set(
            file_set=file_set,
            callable_ref=code_config.callable_ref,
            callable_params=code_config.callable_params,
            install_required=code_config.install_required
        )

    if code_config.file_set is not None \
            and code_config.file_set.is_local_zip():
        # Local ZIP archives will become file-attachments
        # to HTTP service requests. In the request's JSON we will
        # still have the local path.
        # So we are done here.
        return code_config

    if code_config.file_set is not None \
            and code_config.file_set.is_local_zip():
        # At this point, code_config.file_set
        # is always a local ZIP archive.
        return code_config

    raise RuntimeError('for_service() failed due to an '
                       'invalid CodeConfig state')


def _for_local(code_config: CodeConfig) -> 'CodeConfig':
    # noinspection PyProtectedMember
    if code_config._callable is not None:
        # If the callable is already defined,
        # there is no need to do anything else.
        return code_config

    if code_config.inline_code:
        # Turn inline code into a local Python module.
        code_config = _inline_code_to_module(code_config.inline_code,
                                             code_config.callable_ref,
                                             code_config.callable_params)
    file_set = code_config.file_set
    if file_set is None:
        raise RuntimeError('CodeConfig.for_local() failed'
                           ' due to an invalid internal state')

    if file_set.is_local_dir() \
            or (file_set.is_local_zip()
                and not file_set.sub_path):
        # Either a directory or Zip file that can be added
        # to sys.path as is
        return code_config

    # If the file set is not a local directory,
    # turn it into one.
    return code_config.from_file_set(
        file_set=file_set.to_local_dir(),
        callable_ref=code_config.callable_ref,
        install_required=code_config.install_required,
        callable_params=code_config.callable_params
    )


def _load_callable(dir_path: str,
                   callable_ref: str,
                   install_required: bool) -> Callable:
    """
    Load the callable from given *dir_path* using the
    callable reference *callable_red*.
    :param dir_path: A local directory path
        (local ZIPs should work too).
    :param callable_ref: A callable reference of form "module:callable"
    :param install_required: Whether source code in *dir_path*
        must be installed using setup.
    :return: The loaded callable
    """
    module_name, callable_name = _normalize_callable_ref(callable_ref)
    if install_required:
        warnings.warn(f'This user-code configuration requires '
                      f'package installation, '
                      f'but this is not supported yet')
    if dir_path not in sys.path:
        # Ok, we need to change global state here.
        # This cannot be done temporarily (e.g., by using a context manager)
        # as additional code might be dynamically loaded later
        # from local imports.
        LOG.warn(f'Python sys.path prepended by {dir_path}')
        sys.path = [dir_path] + sys.path
    # Now we should be able to import the module...
    # (import_module() will raise an ImportError otherwise).
    module = importlib.import_module(module_name)
    # ...and find the callable.
    _callable = getattr(module, callable_name, None)
    if _callable is None:
        raise ImportError(f'callable {callable_name!r}'
                          f' not found in module {module_name!r}')
    if not callable(_callable):
        raise ImportError(f'{callable_name} of module {callable_name}'
                          f' is not callable')
    LOG.info(f'Imported {callable_ref!r} from {dir_path}')
    return _callable


_user_module_counter = 0


def _next_user_module_name() -> str:
    global _user_module_counter
    _user_module_counter += 1
    return f'{DEFAULT_MODULE_NAME}_{_user_module_counter}'


def _callable_to_module(_callable: Callable,
                        callable_params: Dict[str, Any] = None) -> CodeConfig:
    """
    Create a code configuration from the callable *_callable*.

    This will create a configuration that uses a ``file_set``
    which contains the source code for the *func_or_class*.

    :param _callable: A function or class
    :param callable_params: The parameters passed
        as keyword-arguments to the callable.
    :return: A new code configuration.
    """
    callable_name = _callable.__name__
    if not callable_name:
        raise ValueError(f'cannot detect name '
                         f'for func_or_class')

    module = inspect.getmodule(_callable)
    module_name = module.__name__ if module is not None else None
    if not module_name:
        raise ValueError(f'cannot detect module '
                         f'for callable {callable_name!r}')

    source_file = inspect.getabsfile(_callable)
    if source_file is None:
        raise ValueError(f'cannot detect source file '
                         f'for func_or_class {callable_name!r}')

    module_path, ext = os.path.splitext(os.path.normpath(source_file))
    if not module_path.replace(os.path.sep, '/') \
            .endswith(module_name.replace('.', '/')):
        raise ValueError(f'cannot detect module path '
                         f'for func_or_class {callable_name!r}')

    module_path = os.path.normpath(module_path[0: -len(module_name)])

    code_config = CodeConfig.from_file_set(
        file_set=FileSet(module_path,
                         includes=['*.py']),
        callable_ref=f'{module_name}:{callable_name}',
        callable_params=callable_params
    )
    code_config.set_callable(_callable)
    return code_config


def _inline_code_to_module(inline_code: str,
                           callable_ref: str,
                           callable_params: Dict[str, Any] = None):
    module_name, callable_name = _normalize_callable_ref(callable_ref)
    dir_path = new_temp_dir(prefix=TEMP_FILE_PREFIX)
    with open(os.path.join(dir_path, f'{module_name}.py'), 'w') as stream:
        stream.write(inline_code)
    return CodeConfig.from_file_set(file_set=FileSet(dir_path),
                                    callable_ref=callable_ref,
                                    callable_params=callable_params)


def _normalize_file_set(file_set: Union[FileSet, str, Any]) -> FileSet:
    if isinstance(file_set, FileSet):
        return file_set
    return FileSet(str(file_set))


def _normalize_inline_code(
        code: Tuple[Union[str, Callable], ...],
        callable_name: str = None,
        module_name: str = None
) -> Tuple[str, str]:
    if not callable_name:
        first_callable = next(filter(callable, code), None)
        if first_callable is not None:
            callable_name = first_callable.__name__
    callable_name = callable_name or DEFAULT_CALLABLE_NAME
    module_name = module_name or _next_user_module_name()
    if all(map(callable, code)):
        warnings.warn('inline code may not be executable due to missing '
                      'import statements')
    inline_code = '\n\n'.join([c if isinstance(c, str)
                               else inspect.getsource(c)
                               for c in code])
    return inline_code, f'{module_name}:{callable_name}'


def _normalize_callable_ref(callable_ref: str) -> Tuple[Optional[str], str]:
    """Normalize a callable reference *callable_ref* of the form
    "[module_name:]callable_name" into a pair.
    """
    assert_instance(callable_ref, str, 'callable_ref')

    splits = callable_ref.split(':')
    if len(splits) == 2:
        module_name, callable_name = [s.strip() for s in splits]
    else:
        module_name, callable_name = '', ''

    # Verify components

    assert_true(module_name and callable_name,
                message=f'illegal callable_ref {callable_ref!r}: '
                        f'expected format "<module>:<callable>"')
    assert_true(all(s.isidentifier() for s in module_name.split('.')),
                message=f'illegal callable_ref {callable_ref!r}: '
                        f'{module_name!r} is not a valid Python module name')
    assert_true(callable_name.isidentifier(),
                message=f'illegal callable_ref {callable_ref!r}: '
                        f'{callable_name!r} is not a valid Python identifier')

    return module_name, callable_name



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os.path
import shutil
from typing import Type, Union, Sequence

import zarr

from xcube.core.unchunk import unchunk_dataset


def optimize_dataset(input_path: str,
                     output_path: str = None,
                     in_place: bool = False,
                     unchunk_coords: Union[bool, str, Sequence[str]] = False,
                     exception_type: Type[Exception] = ValueError):
    """
    Optimize a dataset for faster access.

    Reduces the number of metadata and coordinate data files in xcube dataset given by given by *dataset_path*.
    Consolidated cubes open much faster from remote locations, e.g. in object storage,
    because obviously much less HTTP requests are required to fetch initial cube meta
    information. That is, it merges all metadata files into a single top-level JSON file ".zmetadata".

    If *unchunk_coords* is given, it also removes any chunking of coordinate variables
    so they comprise a single binary data file instead of one file per data chunk.
    The primary usage of this function is to optimize data cubes for cloud object storage.
    The function currently works only for data cubes using Zarr format.
    *unchunk_coords* can be a name, or list of names of the coordinate variable(s) to be consolidated.
    If boolean ``True`` is used, coordinate all variables will be consolidated.

    :param input_path: Path to input dataset with ZARR format.
    :param output_path: Path to output dataset with ZARR format. May contain "{input}" template string,
           which is replaced by the input path's file name without file name extension.
    :param in_place: Whether to modify the dataset in place.
           If False, a copy is made and *output_path* must be given.
    :param unchunk_coords: The name of a coordinate variable or a list of coordinate variables whose chunks should
           be consolidated. Pass ``True`` to consolidate chunks of all coordinate variables.
    :param exception_type: Type of exception to be used on value errors.
    """

    if not os.path.isfile(os.path.join(input_path, '.zgroup')):
        raise exception_type('Input path must point to ZARR dataset directory.')

    input_path = os.path.abspath(os.path.normpath(input_path))

    if in_place:
        output_path = input_path
    else:
        if not output_path:
            raise exception_type(f'Output path must be given.')
        if '{input}' in output_path:
            base_name, _ = os.path.splitext(os.path.basename(input_path))
            output_path = output_path.format(input=base_name)
        output_path = os.path.abspath(os.path.normpath(output_path))
        if os.path.exists(output_path):
            raise exception_type(f'Output path already exists.')

    if not in_place:
        shutil.copytree(input_path, output_path)

    if unchunk_coords:
        if isinstance(unchunk_coords, str):
            var_names = (unchunk_coords,)
        elif isinstance(unchunk_coords, bool):
            var_names = None
        else:
            var_names = tuple(unchunk_coords)
        unchunk_dataset(output_path, var_names=var_names, coords_only=True)

    zarr.convenience.consolidate_metadata(output_path)



import warnings
from typing import Type


def register_json_formatter(cls: Type, to_dict_method_name: str = 'to_dict'):
    """
    TODO
    :param cls:
    :param to_dict_method_name:
    :return:
    """
    if not hasattr(cls, to_dict_method_name) or not callable(getattr(cls, to_dict_method_name)):
        raise ValueError(f'{cls} must define a {to_dict_method_name}() method')

    try:
        import IPython
        import IPython.display

        if IPython.get_ipython() is not None:
            def obj_to_dict(obj):
                return getattr(obj, to_dict_method_name)()

            ipy_formatter = IPython.get_ipython().display_formatter.formatters['application/json']
            ipy_formatter.for_type(cls, obj_to_dict)

    except ImportError:
        pass


def enable_asyncio():
    """
    Enable asyncio package to be executable in Jupyter Notebooks.
    """
    try:
        import IPython
        if IPython.get_ipython() is not None:
            try:
                import nest_asyncio
                nest_asyncio.apply()
            except ImportError:
                warnings.warn('nest-asyncio required to use asyncio in Jupyter Notebooks')
    except ImportError:
        pass



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import datetime
import re
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import cftime
import numpy as np
import pandas as pd
import xarray as xr

from xcube.util.assertions import assert_in

REF_DATETIME_STR = '1970-01-01 00:00:00'
REF_DATETIME = pd.to_datetime(REF_DATETIME_STR, utc=True)
DATETIME_UNITS = f'days since {REF_DATETIME_STR}'
DATETIME_CALENDAR = 'gregorian'
SECONDS_PER_DAY = 24 * 60 * 60
MICROSECONDS_PER_DAY = 1000 * 1000 * SECONDS_PER_DAY

_RE_TO_DATETIME_FORMATS = patterns = [(re.compile(14 * '\\d'), '%Y%m%d%H%M%S'),
                                      (re.compile(12 * '\\d'), '%Y%m%d%H%M'),
                                      (re.compile(8 * '\\d'), '%Y%m%d'),
                                      (re.compile(6 * '\\d'), '%Y%m'),
                                      (re.compile(4 * '\\d'), '%Y')]


def add_time_coords(dataset: xr.Dataset,
                    time_range: Tuple[float, float]) -> xr.Dataset:
    t1, t2 = time_range
    if t1 != t2:
        t_center = (t1 + t2) / 2
    else:
        t_center = t1
    dataset = dataset.expand_dims('time')
    dataset = dataset.assign_coords(time=(['time'],
                                          from_time_in_days_since_1970(
                                              [t_center])))
    time_var = dataset.coords['time']
    time_var.attrs['long_name'] = 'time'
    time_var.attrs['standard_name'] = 'time'
    # Avoiding xarray error:
    #   ValueError: failed to prevent overwriting existing key units in attrs on variable 'time'.
    #   This is probably an encoding field used by xarray to describe how a variable is serialized.
    #   To proceed, remove this key from the variable's attributes manually.
    # time_var.attrs['units'] = DATETIME_UNITS
    # time_var.attrs['calendar'] = DATETIME_CALENDAR
    time_var.encoding['units'] = DATETIME_UNITS
    time_var.encoding['calendar'] = DATETIME_CALENDAR
    if t1 != t2:
        time_var.attrs['bounds'] = 'time_bnds'
        dataset = dataset.assign_coords(
            time_bnds=(['time', 'bnds'],
                       from_time_in_days_since_1970([t1, t2]).reshape(1, 2))
        )
        time_bnds_var = dataset.coords['time_bnds']
        time_bnds_var.attrs['long_name'] = 'time'
        time_bnds_var.attrs['standard_name'] = 'time'
        # Avoiding xarray error:
        #   ValueError: failed to prevent overwriting existing key units in attrs on variable
        #   'time'. This is probably an encoding field used by xarray to describe how a variable
        #   is serialized.
        # To proceed, remove this key from the variable's attributes manually.
        # time_bnds_var.attrs['units'] = DATETIME_UNITS
        # time_bnds_var.attrs['calendar'] = DATETIME_CALENDAR
        time_bnds_var.encoding['units'] = DATETIME_UNITS
        time_bnds_var.encoding['calendar'] = DATETIME_CALENDAR
    return dataset


def get_time_range_from_data(dataset: xr.Dataset,
                             maybe_consider_metadata: bool = True) \
        -> Tuple[Optional[float], Optional[float]]:
    """
    Determines a time range from a dataset by inspecting its time_bounds or time data arrays.
    In cases where no time bounds are given and no time periodicity can be determined,
    metadata may be considered.

    :param dataset: The dataset of which the time range shall be determined
    :param maybe_consider_metadata: Whether metadata shall be considered.
    Only used when the dataset has no time bounds array and no time periodicity.
    The values will only be set when they do not contradict the values from the data arrays.
    :return: A tuple with two float values: The first one represents the start time,
    the second the end time. Either may be None.
    """
    time_bounds_names = ['time_bnds', 'time_bounds']
    for time_bounds_name in time_bounds_names:
        if time_bounds_name in dataset:
            return _get_time_range_from_time_bounds(dataset, time_bounds_name)
    # handle special case with datasets that have special coordinates 'start_time' and 'end_time'.
    # This is the case for, e.g, ICESHEETS Greenland
    if 'start_time' in dataset and 'end_time' in dataset:
        return dataset['start_time'].values[0], dataset['end_time'].values[-1]
    time_names = ['time', 't']
    time = None
    for time_name in time_names:
        if time_name in dataset:
            time = dataset[time_name]
            break
    if time is None:
        return None, None
    time_bnds_name = time.attrs.get("bounds", "time_bnds")
    if time_bnds_name in dataset:
        return _get_time_range_from_time_bounds(dataset, time_bnds_name)
    is_cf_time = isinstance(time[0].values.item(), cftime.datetime)
    data_start = pd.to_datetime(time[0].values.item().isoformat()) \
        if is_cf_time else time[0].values
    data_end = pd.to_datetime(time[-1].values.item().isoformat()) \
        if is_cf_time else time[-1].values
    if time.size < 3:
        return _maybe_return_time_range_from_metadata(dataset,
                                                      data_start,
                                                      data_end,
                                                      maybe_consider_metadata)
    time_diff = time.diff(dim=time.dims[0]).values
    time_res = time_diff[0]
    time_regular = all(
        [time_res - diff == np.timedelta64(0) for diff in time_diff[1:]])
    if time_regular:
        return data_start - time_res / 2, data_end + time_res / 2
    return _maybe_return_time_range_from_metadata(dataset,
                                                  data_start,
                                                  data_end,
                                                  maybe_consider_metadata)


def _maybe_return_time_range_from_metadata(dataset: xr.Dataset,
                                           data_start_time: float,
                                           data_end_time: float,
                                           maybe_consider_metadata: bool) -> \
Tuple[float, float]:
    if maybe_consider_metadata:
        attr_start_time, attr_end_time = get_time_range_from_attrs(dataset)
        attr_start_time = pd.to_datetime(attr_start_time,
                                         infer_datetime_format=False, utc=True)
        attr_end_time = pd.to_datetime(attr_end_time,
                                       infer_datetime_format=False, utc=True)
        if attr_start_time is not None and attr_end_time is not None:
            try:
                if attr_start_time < data_start_time and attr_end_time > data_end_time:
                    return attr_start_time.to_datetime64(), attr_end_time.to_datetime64()
            except TypeError:
                try:
                    if attr_start_time.to_datetime64() < data_start_time \
                            and attr_end_time.to_datetime64() > data_end_time:
                        return attr_start_time.to_datetime64(), attr_end_time.to_datetime64()
                except TypeError:
                    # use time values from data
                    pass
    return data_start_time, data_end_time


def _get_time_range_from_time_bounds(dataset: xr.Dataset, time_bounds_name: str) \
        -> Tuple[Optional[float], Optional[float]]:
    time_bnds = dataset[time_bounds_name]
    if len(time_bnds.shape) == 2 and time_bnds.shape[1] == 2:
        return time_bnds[0, 0].values, time_bnds[-1, 1].values


def get_time_range_from_attrs(dataset: xr.Dataset) -> Tuple[
    Optional[str], Optional[str]]:
    return get_start_time_from_attrs(dataset), get_end_time_from_attrs(dataset)


def get_start_time_from_attrs(dataset: xr.Dataset) -> Optional[str]:
    return _get_attr(dataset,
                     ['time_coverage_start', 'time_start', 'start_time',
                      'start_date'])


def get_end_time_from_attrs(dataset: xr.Dataset) -> Optional[str]:
    return _get_attr(dataset,
                     ['time_coverage_end', 'time_stop', 'time_end', 'stop_time',
                      'end_time', 'stop_date', 'end_date'])


def _get_attr(dataset: xr.Dataset, names: Sequence[str]) -> Optional[str]:
    for name in names:
        if name in dataset.attrs:
            return remove_time_part_from_isoformat(str(dataset.attrs[name]))


def remove_time_part_from_isoformat(datetime_str: str) -> str:
    date_length = 10  # for example len("2010-02-04") == 10
    if len(datetime_str) > date_length and datetime_str[date_length] in (
    'T', ' '):
        return datetime_str[0: date_length]
    return datetime_str


def to_time_in_days_since_1970(time_str: str, pattern=None) -> float:
    date_time = pd.to_datetime(time_str, format=pattern,
                               infer_datetime_format=False, utc=True)
    timedelta = date_time - REF_DATETIME
    return timedelta.days + timedelta.seconds / SECONDS_PER_DAY + \
        timedelta.microseconds / MICROSECONDS_PER_DAY


def from_time_in_days_since_1970(
        time_value: Union[float, Sequence[float]]) -> np.ndarray:
    if isinstance(time_value, int) or isinstance(time_value, float):
        return pd.to_datetime(time_value, utc=True, unit='D',
                              origin='unix').round(freq='ms') \
            .to_datetime64()
    else:
        return np.array(list(map(from_time_in_days_since_1970, time_value)))


def timestamp_to_iso_string(time: Union[np.datetime64, datetime.datetime],
                            freq: str = 'S',
                            round_fn: str = 'round'):
    """
    Convert a UTC timestamp given as nanos, millis, seconds, etc. since 1970-01-01 00:00:00
    to an ISO-format string.

    :param time: UTC timestamp given as time delta since 1970-01-01 00:00:00 in the units
        given by the numpy datetime64 type, so it can be as nanos, millis,
        seconds since 1970-01-01 00:00:00.
    :param freq: time rounding resolution. See pandas.Timestamp.round().
    :param round_fn: time rounding function. Defaults to pandas.Timestamp.round().
    :return: ISO-format string.
    """
    # All times are UTC (Z = Zulu Time Zone = UTC)
    assert_in(round_fn, ("ceil", "floor", "round"), name="round_fn")
    timestamp = pd.Timestamp(time)
    return getattr(timestamp, round_fn)(freq).isoformat() + 'Z'


def find_datetime_format(line: str) -> Tuple[Optional[str], int, int]:
    for regex, time_format in _RE_TO_DATETIME_FORMATS:
        searcher = regex.search(line)
        if searcher:
            p1, p2 = searcher.span()
            return time_format, p1, p2
    return None, -1, -1


def get_timestamp_from_string(string: str) -> Optional[pd.Timestamp]:
    time_format, p1, p2 = find_datetime_format(string)
    if time_format:
        try:
            return pd.to_datetime(string[p1:p2], format=time_format)
        except ValueError:
            pass


def get_timestamps_from_string(string: str) -> (pd.Timestamp, pd.Timestamp):
    first_time = None
    second_time = None
    time_format, p1, p2 = find_datetime_format(string)
    try:
        if time_format:
            first_time = pd.to_datetime(string[p1:p2], format=time_format)
        string_rest = string[p2:]
        time_format, p1, p2 = find_datetime_format(string_rest)
        if time_format:
            second_time = pd.to_datetime(string_rest[p1:p2], format=time_format)
    except ValueError:
        pass
    return first_time, second_time



import unittest

from xcube.plugin import init_plugin
from xcube.util.extension import ExtensionRegistry


class PluginTest(unittest.TestCase):
    def test_standard_extensions_exist(self):
        ext_reg = ExtensionRegistry()

        init_plugin(ext_reg)

        self.assertTrue(ext_reg.has_extension('xcube.core.gen.iproc', 'default'))

        self.assertTrue(ext_reg.has_extension('xcube.core.dsio', 'zarr'))
        self.assertTrue(ext_reg.has_extension('xcube.core.dsio', 'netcdf4'))
        self.assertTrue(ext_reg.has_extension('xcube.core.dsio', 'csv'))
        self.assertTrue(ext_reg.has_extension('xcube.core.dsio', 'mem'))

        self.assertTrue(ext_reg.has_extension('xcube.core.store', 'file'))
        self.assertTrue(ext_reg.has_extension('xcube.core.store', 'memory'))
        self.assertTrue(ext_reg.has_extension('xcube.core.store', 's3'))

        self.assertTrue(ext_reg.has_extension('xcube.cli', 'compute'))
        self.assertTrue(ext_reg.has_extension('xcube.cli', 'extract'))
        self.assertTrue(ext_reg.has_extension('xcube.cli', 'gen'))
        self.assertTrue(ext_reg.has_extension('xcube.cli', 'level'))
        self.assertTrue(ext_reg.has_extension('xcube.cli', 'optimize'))




# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pyproj.crs
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.resampling import resample_in_space
from .transformer import CubeTransformer
from .transformer import TransformedCube
from ..config import CubeConfig


class CubeResamplerXY(CubeTransformer):

    def transform_cube(self,
                       cube: xr.Dataset,
                       source_gm: GridMapping,
                       cube_config: CubeConfig) -> TransformedCube:
        target_gm = _compute_target_grid_mapping(cube_config, source_gm)

        if source_gm.is_close(target_gm):
            resampled_cube = cube
        else:
            resampled_cube = resample_in_space(
                cube,
                source_gm=source_gm,
                target_gm=target_gm
            )

        cube_config = cube_config.drop_props(['crs', 'spatial_res', 'bbox'])

        return resampled_cube, target_gm, cube_config


def _compute_target_grid_mapping(cube_config: CubeConfig,
                                 source_gm: GridMapping) -> GridMapping:
    # assert_true(source_gm.is_regular, 'source_gm must be regular')

    target_crs = cube_config.crs
    target_bbox = cube_config.bbox
    target_spatial_res = cube_config.spatial_res

    if target_crs is None \
            and target_bbox is None \
            and target_spatial_res is None:
        # Nothing to do
        if source_gm.is_regular:
            return source_gm
        return source_gm.to_regular(tile_size=cube_config.tile_size)

    if target_spatial_res is not None:
        xy_res = (target_spatial_res, target_spatial_res)
    else:
        xy_res = source_gm.xy_res
    if target_bbox is not None:
        x_res, y_res = xy_res
        x_min, y_min, x_max, y_max = target_bbox
        xy_min = x_min, y_min
        size = round((x_max - x_min) / x_res), round((y_max - y_min) / y_res)
    else:
        xy_min = source_gm.x_min, source_gm.y_min
        size = source_gm.size
    if target_crs is not None:
        crs = pyproj.crs.CRS.from_string(target_crs)
    else:
        crs = source_gm.crs
    target_gm = GridMapping.regular(size=size,
                                    xy_min=xy_min,
                                    xy_res=xy_res,
                                    crs=crs,
                                    tile_size=source_gm.tile_size,
                                    is_j_axis_up=source_gm.is_j_axis_up)
    return target_gm.derive(xy_var_names=source_gm.xy_var_names,
                            xy_dim_names=source_gm.xy_dim_names)



# The MIT License (MIT)
# Copyright (c) 2019-2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Dict, Any, Iterable

import dask.array as da
import numpy as np
import xarray as xr


# TODO: this would be useful to have in xarray:
#       >>> ds = xr.open_dataset("my/path/to/cf/netcdf.nc", decode_flags=True)
#       >>> ds.flag_mask_sets['quality_flags']

class MaskSet:
    """
    A set of mask variables derived from a variable *flag_var* with the following
    CF attributes:

      - One or both of `flag_masks` and `flag_values`
      - `flag_meanings` (always required)

    See https://cfconventions.org/Data/cf-conventions/cf-conventions-1.9/cf-conventions.html#flags
    for details on the use of these attributes.

    Each mask is represented by an `xarray.DataArray`, has the name of the flag,
    is of type `numpy.unit8`, and has the dimensions of the given *flag_var*.

    :param flag_var: an `xarray.DataArray` that defines flag values.
           The CF attributes `flag_meanings` and one or both of
           `flag_masks` and `flag_values` are expected to exist and be valid.
    """

    def __init__(self, flag_var: xr.DataArray):
        flag_masks = flag_var.attrs.get('flag_masks')
        flag_values = flag_var.attrs.get('flag_values')
        if flag_masks is None and flag_values is None:
            raise ValueError("One or both of the attributes "
                             "'flag_masks' or 'flag_values' "
                             "must be present and non-null in flag_var")
        if flag_masks is not None:
            flag_masks = _convert_flag_var_attribute_value(flag_masks,
                                                           'flag_masks')
        if flag_values is not None:
            flag_values = _convert_flag_var_attribute_value(flag_values,
                                                            'flag_values')
        if 'flag_meanings' not in flag_var.attrs:
            raise ValueError(
                "flag_var must have the attribute 'flag_meanings'")
        flag_meanings = flag_var.attrs.get('flag_meanings')
        if not isinstance(flag_meanings, str):
            raise TypeError("attribute 'flag_meanings' of flag_var "
                            "must be a string")
        flag_names = flag_meanings.split(' ')
        if flag_masks is not None and len(flag_names) != len(flag_masks):
            raise ValueError("attributes 'flag_meanings' and 'flag_masks' "
                             "are not corresponding")
        if flag_values is not None and len(flag_names) != len(flag_values):
            raise ValueError("attributes 'flag_meanings' and 'flag_values' "
                             "are not corresponding")
        if flag_masks is None:
            flag_masks = [None] * len(flag_names)
        if flag_values is None:
            flag_values = [None] * len(flag_names)
        self._flag_var = flag_var
        self._flag_names = flag_names
        self._flags = dict(zip(flag_names, list(zip(flag_masks, flag_values))))
        self._masks = {}

    @classmethod
    def is_flag_var(cls, var: xr.DataArray) -> bool:
        return 'flag_meanings' in var.attrs \
               and ('flag_masks' in var.attrs or 'flag_values' in var.attrs)

    @classmethod
    def get_mask_sets(cls, dataset: xr.Dataset) -> Dict[str, 'MaskSet']:
        """
        For each "flag" variable in given *dataset*, turn it into a ``MaskSet``,
        store it in a dictionary.

        :param dataset: The dataset
        :return: A mapping of flag names to ``MaskSet``.
            Will be empty if there are no flag variables in *dataset*.
        """
        masks = {}
        for var_name in dataset.variables:
            var = dataset[var_name]
            if cls.is_flag_var(var):
                masks[var_name] = MaskSet(var)
        return masks

    def _repr_html_(self):
        lines = ['<html>',
                 '<table>',
                 '<tr><th>Flag name</th><th>Mask</th><th>Value</th></tr>']

        for name, data in self._flags.items():
            mask, value = data
            lines.append(f'<tr><td>{name}</td><td>{mask}</td><td>{value}</td></tr>')

        lines.extend(['</table>',
                      '</html>'])

        return '\n'.join(lines)

    def __str__(self):
        return "%s(%s)" % (self._flag_var.name, ', '.join(["%s=%s" % (n, v) for n, v in self._flags.items()]))

    def __dir__(self) -> Iterable[str]:
        return self._flag_names

    def __getattr__(self, name: str) -> Any:
        if name not in self._flags:
            raise AttributeError(name)
        return self.get_mask(name)

    def __getitem__(self, item):
        try:
            name = self._flag_names[item]
            if name not in self._flags:
                raise IndexError(item)
        except TypeError:
            name = item
            if name not in self._flags:
                raise KeyError(item)
        return self.get_mask(name)

    def __contains__(self, item):
        return item in self._flags

    def get_mask(self, flag_name: str):
        if flag_name not in self._flags:
            raise ValueError('invalid flag name "%s"' % flag_name)

        if flag_name in self._masks:
            return self._masks[flag_name]

        flag_var = self._flag_var
        flag_mask, flag_value = self._flags[flag_name]

        if flag_var.chunks is not None:
            ones_array = da.ones(flag_var.shape,
                                 dtype=np.uint8,
                                 chunks=flag_var.chunks)
        else:
            ones_array = np.ones(flag_var.shape,
                                 dtype=np.uint8)

        mask_var = xr.DataArray(ones_array,
                                dims=flag_var.dims,
                                name=flag_name,
                                coords=flag_var.coords)
        if flag_mask is not None:
            if flag_var.dtype != flag_mask.dtype:
                flag_var = flag_var.astype(flag_mask.dtype)
            if flag_value is not None:
                mask_var = mask_var.where((flag_var & flag_mask) == flag_value, 0)
            else:
                mask_var = mask_var.where((flag_var & flag_mask) != 0, 0)
        else:
            if flag_var.dtype != flag_value.dtype:
                flag_var = flag_var.astype(flag_value.dtype)
            mask_var = mask_var.where(flag_var == flag_value, 0)

        self._masks[flag_name] = mask_var
        return mask_var


_MASK_DTYPES = (
    (2 ** 8, np.uint8), (2 ** 16, np.uint16), (2 ** 32, np.uint32), (2 ** 64, np.uint64)
)


def _convert_flag_var_attribute_value(attr_value, attr_name):
    if isinstance(attr_value, str):
        err_msg = f'Invalid bit expression in value for {attr_name}: "{attr_value}"'
        masks = []
        max_mask = 0
        for s in attr_value.split(","):
            s = s.strip()
            pair = s.split('-')
            if len(pair) == 1:
                try:
                    mask = (1 << int(s[0:-1])) if s.endswith("b") else int(s)
                except ValueError as e:
                    raise ValueError(err_msg) from e
            elif len(pair) == 2:
                s1, s2 = pair
                if not s1.endswith("b") or not s2.endswith("b"):
                    raise ValueError(err_msg)
                try:
                    b1 = int(s1[0:-1])
                    b2 = int(s2[0:-1])
                except ValueError as e:
                    raise ValueError(err_msg) from e
                if b1 > b2:
                    raise ValueError(err_msg)
                mask = 0
                for b in range(b1, b2 + 1):
                    mask |= 1 << b
            else:
                raise ValueError(err_msg)
            masks.append(mask)
            max_mask = max(max_mask, mask)

        for limit, dtype in _MASK_DTYPES:
            if max_mask <= limit:
                return np.array(masks, dtype)

        raise ValueError(err_msg)

    if not (hasattr(attr_value, 'dtype') and hasattr(attr_value, 'shape')):
        raise TypeError(f'attribute {attr_name!r} must be an integer array')

    return attr_value



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Union, Callable, Mapping, Hashable, Any, Optional

import numpy as np
import xarray as xr
from dask import array as da

from xcube.core.gridmapping import GridMapping
from xcube.core.gridmapping.helpers import scale_xy_res_and_size
from .affine import affine_transform_dataset
from .affine import resample_dataset
from .rectify import rectify_dataset

NDImage = Union[np.ndarray, da.Array]
Aggregator = Callable[[NDImage], NDImage]

# If _SCALE_LIMIT is exceeded, we don't need
# to downscale source image before we can
# rectify it.
_SCALE_LIMIT = 0.95


def resample_in_space(
        dataset: xr.Dataset,
        source_gm: GridMapping = None,
        target_gm: GridMapping = None,
        var_configs: Mapping[Hashable, Mapping[str, Any]] = None,
        encode_cf: bool = True,
        gm_name: Optional[str] = None
):
    """
    Resample a dataset in the spatial dimensions.

    If the source grid mapping *source_gm* is not given,
    it is derived from *dataset*:
    ``source_gm = GridMapping.from_dataset(dataset)``.

    If the target grid mapping *target_gm* is not given,
    it is derived from *source_gm*:
    ``target_gm = source_gm.to_regular()``.

    If *source_gm* is almost equal to *target_gm*, this
    function is a no-op and *dataset* is returned unchanged.

    Otherwise the function computes a spatially
    resampled version of *dataset* and returns it.

    Using *var_configs*, the resampling of individual
    variables can be configured. If given, *var_configs*
    must be a mapping from variable names to configuration
    dictionaries which can have the following properties:

    * ``spline_order`` (int) - The order of spline polynomials
        used for interpolating. It is used for upsampling only.
        Possible values are 0 to 5.
        Default is 1 (bi-linear) for floating point variables,
        and 0 (= nearest neighbor) for integer and bool variables.
    * ``aggregator`` (str) - An optional aggregating
        function. It is used for downsampling only.
        Examples are numpy.nanmean, numpy.nanmin, numpy.nanmax.
        Default is numpy.nanmean for floating point variables,
        and None (= nearest neighbor) for integer and bool variables.
    * ``recover_nan`` (bool) - whether a special algorithm
        shall be used that is able to recover values that would
        otherwise yield NaN during resampling.
        Default is True for floating point variables,
        and False for integer and bool variables.

    Note that *var_configs* is only used if the resampling involves
    an affine transformation. This is true if the CRS of
    *source_gm* and CRS of *target_gm* are equal and one of two
    cases is given:

    1. *source_gm* is regular.
       In this case the resampling is the affine transformation.
       and the result is returned directly.
    2. *source_gm* is not regular and has a lower resolution
       than *target_cm*.
       In this case *dataset* is downsampled first using an affine
       transformation. Then the result is rectified.

    In all other cases, no affine transformation is applied and
    the resampling is a direct rectification.

    :param dataset: The source dataset.
    :param source_gm: The source grid mapping.
    :param target_gm: The target grid mapping. Must be regular.
    :param var_configs: Optional resampling configurations
        for individual variables.
    :param encode_cf: Whether to encode the target grid mapping
        into the resampled dataset in a CF-compliant way.
        Defaults to ``True``.
    :param gm_name: Name for the grid mapping variable.
        Defaults to "crs". Used only if *encode_cf* is ``True``.
    :return: The spatially resampled dataset.
    """
    if source_gm is None:
        # No source grid mapping given, so do derive it from dataset
        source_gm = GridMapping.from_dataset(dataset)

    if target_gm is None:
        # No target grid mapping given, so do derive it from source
        target_gm = source_gm.to_regular()

    if source_gm.is_close(target_gm):
        # If source and target grid mappings are almost equal
        return dataset

    # target_gm must be regular
    GridMapping.assert_regular(target_gm, name='target_gm')

    # Are source and target both geographic grid mappings?
    both_geographic = source_gm.crs.is_geographic \
                      and target_gm.crs.is_geographic

    if both_geographic or source_gm.crs == target_gm.crs:
        # If CRSes are both geographic or their CRSes are equal:
        if source_gm.is_regular:
            # If also the source is regular, then resampling reduces
            # to an affine transformation.
            return affine_transform_dataset(
                dataset,
                source_gm=source_gm,
                target_gm=target_gm,
                var_configs=var_configs,
                encode_cf=encode_cf,
                gm_name=gm_name
            )

        # If the source is not regular, we need to rectify it,
        # so the target is regular. Our rectification implementation
        # works only correctly if source pixel size >= target pixel
        # size. Therefore check if we must downscale source first.
        x_scale = source_gm.x_res / target_gm.x_res
        y_scale = source_gm.y_res / target_gm.y_res
        if x_scale > _SCALE_LIMIT and y_scale > _SCALE_LIMIT:
            # Source pixel size >= target pixel size.
            # We can rectify.
            return rectify_dataset(
                dataset,
                source_gm=source_gm,
                target_gm=target_gm,
                encode_cf=encode_cf,
                gm_name=gm_name
            )

        # Source has higher resolution than target.
        # Downscale first, then rectify
        if source_gm.is_regular:
            # If source is regular
            downscaled_gm = source_gm.scale((x_scale, y_scale))
            downscaled_dataset = resample_dataset(
                dataset,
                ((x_scale, 1, 0), (1, y_scale, 0)),
                size=downscaled_gm.size,
                tile_size=source_gm.tile_size,
                xy_dim_names=source_gm.xy_dim_names,
                var_configs=var_configs,
            )
        else:
            _, downscaled_size = scale_xy_res_and_size(source_gm.xy_res,
                                                       source_gm.size,
                                                       (x_scale, y_scale))
            downscaled_dataset = resample_dataset(
                dataset,
                ((x_scale, 1, 0), (1, y_scale, 0)),
                size=downscaled_size,
                tile_size=source_gm.tile_size,
                xy_dim_names=source_gm.xy_dim_names,
                var_configs=var_configs,
            )
            downscaled_gm = GridMapping.from_dataset(
                downscaled_dataset,
                tile_size=source_gm.tile_size,
                prefer_crs=source_gm.crs
            )
        return rectify_dataset(downscaled_dataset,
                               source_gm=downscaled_gm,
                               target_gm=target_gm,
                               encode_cf=encode_cf,
                               gm_name=gm_name)

    # If CRSes are not both geographic and their CRSes are different
    # transform the source_gm so its CRS matches the target CRS:
    transformed_source_gm = source_gm.transform(target_gm.crs)
    transformed_x, transformed_y = transformed_source_gm.xy_coords
    return resample_in_space(
        dataset.assign(transformed_x=transformed_x,
                       transformed_y=transformed_y),
        source_gm=transformed_source_gm,
        target_gm=target_gm,
        gm_name=gm_name
    )



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Sequence, Any, Dict, Callable, Optional

import xarray as xr

from .abc import MultiLevelDataset
from .lazy import LazyMultiLevelDataset


class CombinedMultiLevelDataset(LazyMultiLevelDataset):
    """A multi-level dataset that is a combination of other
    multi-level datasets.

    :param ml_datasets: The multi-level datasets to be combined.
        At least two must be provided.
    :param ds_id: Optional dataset identifier.
    :param combiner_function: An optional function used to combine the
        datasets, for example ``xarray.merge``.
        If given, it receives a list of datasets
        (``xarray.Dataset`` instances) and *combiner_params* as keyword
        arguments.
        If not given or ``None`` is passed, a copy of the first dataset
        is made, which is then subsequently updated by the remaining datasets
        using ``xarray.Dataset.update()``.
    :param combiner_params: Parameters to the *combiner_function*
        passed as keyword arguments.
    """

    def __init__(self,
                 ml_datasets: Sequence[MultiLevelDataset],
                 ds_id: Optional[str] = None,
                 combiner_function: Optional[Callable] = None,
                 combiner_params: Optional[Dict[str, Any]] = None):
        if not ml_datasets or len(ml_datasets) < 2:
            raise ValueError('ml_datasets must have at least two elements')
        super().__init__(ds_id=ds_id,
                         parameters=combiner_params)
        self._ml_datasets = ml_datasets
        self._combiner_function = combiner_function

    def _get_num_levels_lazily(self) -> int:
        return self._ml_datasets[0].num_levels

    def _get_dataset_lazily(self, index: int,
                            combiner_params: Dict[str, Any]) -> xr.Dataset:
        datasets = [ml_dataset.get_dataset(index)
                    for ml_dataset in self._ml_datasets]
        if self._combiner_function is None:
            combined_dataset = datasets[0].copy()
            for dataset in datasets[1:]:
                combined_dataset.update(dataset)
            return combined_dataset
        else:
            return self._combiner_function(datasets, **combiner_params)

    def close(self):
        for ml_dataset in self._ml_datasets:
            ml_dataset.close()



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import fnmatch
import os.path
import pathlib
import uuid
import warnings
from threading import RLock
from typing import Optional, Iterator, Any, Tuple, List, Dict, \
    Union, Container, Callable, Sequence

import fsspec
import geopandas as gpd
import xarray as xr

from xcube.core.mldataset import MultiLevelDataset
from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_in
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.extension import Extension
from xcube.util.fspath import is_local_fs
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonComplexSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNullSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from .accessor import FsAccessor
from .accessor import STORAGE_OPTIONS_PARAM_NAME
from ..accessor import DataOpener
from ..accessor import DataWriter
from ..accessor import find_data_opener_extensions
from ..accessor import find_data_writer_extensions
from ..accessor import get_data_accessor_predicate
from ..accessor import new_data_opener
from ..accessor import new_data_writer
from ..assertions import assert_valid_params
from ..datatype import ANY_TYPE
from ..datatype import DATASET_TYPE
from ..datatype import DataType
from ..datatype import DataTypeLike
from ..datatype import GEO_DATA_FRAME_TYPE
from ..datatype import MULTI_LEVEL_DATASET_TYPE
from ..descriptor import DataDescriptor
from ..descriptor import new_data_descriptor
from ..error import DataStoreError
from ..search import DefaultSearchMixin
from ..store import MutableDataStore

_DEFAULT_DATA_TYPE = DATASET_TYPE.alias
_DEFAULT_FORMAT_ID = 'zarr'

# TODO (forman): The following constants _FILENAME_EXT_TO_DATA_TYPE_ALIASES
#   and _FILENAME_EXT_TO_FORMAT reflect implicit knowledge about the
#   implemented accessor extensions. Let every accessor also provide
#   its allowed file extensions. Then this information can be generated
#   from all registered accessors.

_FILENAME_EXT_TO_FORMAT = {
    '.zarr': 'zarr',
    '.levels': 'levels',
    '.nc': 'netcdf',
    '.tif': 'geotiff',
    '.tiff': 'geotiff',
    '.geotiff': 'geotiff',
    '.shp': 'shapefile',
    '.geojson': 'geojson',
}

_FORMAT_TO_DATA_TYPE_ALIASES = {
    'zarr': (DATASET_TYPE.alias,),
    'netcdf': (DATASET_TYPE.alias,),
    'levels': (MULTI_LEVEL_DATASET_TYPE.alias, DATASET_TYPE.alias),
    'geotiff': (MULTI_LEVEL_DATASET_TYPE.alias, DATASET_TYPE.alias),
    'geojson': (GEO_DATA_FRAME_TYPE.alias,),
    'shapefile': (GEO_DATA_FRAME_TYPE.alias,),
}

_DataId = str
_DataIdTuple = Tuple[_DataId, Dict[str, Any]]
_DataIdIter = Iterator[_DataId]
_DataIdTupleIter = Iterator[_DataIdTuple]
_DataIds = Union[_DataIdIter, _DataIdTupleIter]


class BaseFsDataStore(DefaultSearchMixin, MutableDataStore):
    """
    Base class for data stores that use an underlying filesystem
    of type ``fsspec.AbstractFileSystem``.

    The data store is capable of filtering the data identifiers reported
    by ``get_data_ids()``. For this purpose the optional keywords
    `excludes` and `includes` are used which can both take the form of
    a wildcard pattern or a sequence of wildcard patterns:

    * ``excludes``: if given and if any pattern matches the identifier,
      the identifier is not reported.
    * ``includes``: if not given or if any pattern matches the identifier,
      the identifier is reported.

    :param fs: Optional filesystem. If not given,
        :meth:_load_fs() must return a filesystem instance.
    :param root: Root or base directory.
        Defaults to "".
    :param max_depth: Maximum recursion depth. None means limitless.
        Defaults to 1.
    :param read_only: Whether this is a read-only store.
        Defaults to False.
    :param includes: Optional sequence of wildcards that identify included
        filesystem paths. Affects the data identifiers (paths)
        returned by `get_data_ids()`. By default, all paths are included.
    :param excludes: Optional sequence of wildcards that identify excluded
        filesystem paths. Affects the data identifiers (paths)
        returned by `get_data_ids()`. By default, no paths are excluded.
    """

    def __init__(self,
                 fs: Optional[fsspec.AbstractFileSystem] = None,
                 root: str = '',
                 max_depth: Optional[int] = 1,
                 read_only: bool = False,
                 includes: Optional[Union[str, Sequence[str]]] = None,
                 excludes: Optional[Union[str, Sequence[str]]] = None):
        if fs is not None:
            assert_instance(fs, fsspec.AbstractFileSystem, name='fs')
        self._fs = fs
        self._raw_root: str = root or ''
        self._root: Optional[str] = None
        self._max_depth = max_depth
        self._read_only = read_only
        self._includes = self._normalize_wc(includes)
        self._excludes = self._normalize_wc(excludes)
        self._lock = RLock()

    @property
    def protocol(self) -> str:
        """
        Get the filesystem protocol.
        Should be overridden by clients, because the default
        implementation instantiates the filesystem.
        """
        protocol = self.fs.protocol
        return protocol if isinstance(protocol, str) else protocol[0]

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        """
        Return the underlying filesystem .
        :return: An instance of ``fsspec.AbstractFileSystem``.
        """
        if self._fs is None:
            # Lazily instantiate filesystem.
            with self._lock:
                self._fs = self._load_fs()
        return self._fs

    def _load_fs(self) -> fsspec.AbstractFileSystem:
        """
        Get an instance of the underlying filesystem.
        Default implementation raises NotImplementedError.
        :return: An instance of ``fsspec.AbstractFileSystem``.
        """
        raise NotImplementedError('unknown filesystem')

    @property
    def root(self) -> str:
        if self._root is None:
            # Lazily instantiate root path.
            is_local = is_local_fs(self.fs)
            with self._lock:
                root = self._raw_root
                if is_local:
                    root = os.path.abspath(root)
                self._root = pathlib.Path(root).as_posix()
        return self._root

    @property
    def max_depth(self) -> Optional[int]:
        """Maximum recursion depth. None means limitless."""
        return self._max_depth

    @property
    def read_only(self) -> bool:
        """Whether this is a read-only store."""
        return self._read_only

    @property
    def includes(self) -> Tuple[str]:
        """Wildcard patterns that include paths."""
        return self._includes

    @property
    def excludes(self) -> Tuple[str]:
        """Wildcard patterns that exclude paths."""
        return self._excludes

    #########################################################################
    # MutableDataStore impl.

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        wc_schema = JsonComplexSchema(one_of=[
            JsonNullSchema(),
            JsonStringSchema(),
            JsonArraySchema(items=JsonStringSchema())
        ])
        return JsonObjectSchema(
            properties=dict(
                root=JsonStringSchema(default=''),
                max_depth=JsonIntegerSchema(nullable=True, default=1),
                read_only=JsonBooleanSchema(default=False),
                includes=wc_schema,
                excludes=wc_schema,
            ),
            additional_properties=False
        )

    @classmethod
    def get_data_types(cls) -> Tuple[str, ...]:
        return tuple(set(
            data_type
            for types_tuple in _FORMAT_TO_DATA_TYPE_ALIASES.values()
            for data_type in types_tuple
        ))

    def get_data_types_for_data(self, data_id: str) -> Tuple[str, ...]:
        self._assert_valid_data_id(data_id)
        data_type_alias, format_id, protocol = \
            self._guess_accessor_id_parts(data_id)
        data_type_aliases = [data_type_alias]
        for ext in find_data_opener_extensions(get_data_accessor_predicate(
            format_id=format_id,
            storage_id=protocol
        )):
            data_type_alias = ext.name.split(':')[0]
            if data_type_alias not in data_type_aliases:
                data_type_aliases.append(data_type_alias)
        return tuple(data_type_aliases)

    def get_data_ids(self,
                     data_type: DataTypeLike = None,
                     include_attrs: Container[str] = None) -> _DataIds:
        data_type = DataType.normalize(data_type)
        # TODO: do not ignore names in include_attrs
        return_tuples = include_attrs is not None
        data_ids = self._generate_data_ids('', data_type, return_tuples, 1)
        if self._includes or self._excludes:
            yield from self._filter_data_ids(data_ids, return_tuples)
        yield from data_ids

    def has_data(self, data_id: str, data_type: DataTypeLike = None) -> bool:
        assert_given(data_id, 'data_id')
        if self._is_data_specified(data_id, data_type):
            fs_path = self._convert_data_id_into_fs_path(data_id)
            return self.fs.exists(fs_path)
        return False

    def describe_data(self, data_id: str, data_type: DataTypeLike = None) \
            -> DataDescriptor:
        self._assert_valid_data_id(data_id)
        self._assert_data_specified(data_id, data_type)
        # TODO: optimize me, self.open_data() may be very slow!
        #   For Zarr, try using self.fs to load metadata only
        #   rather than instantiating xr.Dataset instances which
        #   can be very expensive for large Zarrs (xarray 0.18.2),
        #   especially in S3 filesystems.
        data = self.open_data(data_id)
        return new_data_descriptor(data_id, data, require=True)

    def get_data_opener_ids(self,
                            data_id: str = None,
                            data_type: DataTypeLike = None) \
            -> Tuple[str, ...]:
        data_type = DataType.normalize(data_type)
        format_id = None
        storage_id = self.protocol
        if data_id:
            accessor_id_parts = self._guess_accessor_id_parts(
                data_id, require=False
            )
            if not accessor_id_parts:
                return ()  # nothing found
            acc_data_type_alias, format_id, storage_id = accessor_id_parts
            if data_type == ANY_TYPE:
                data_type = DataType.normalize(acc_data_type_alias)
        return tuple(ext.name for ext in find_data_opener_extensions(
            predicate=get_data_accessor_predicate(
                data_type=data_type,
                format_id=format_id,
                storage_id=storage_id
            )
        ))

    def get_open_data_params_schema(self,
                                    data_id: str = None,
                                    opener_id: str = None) \
            -> JsonObjectSchema:
        opener = self._find_opener(opener_id=opener_id, data_id=data_id)
        return self._get_open_data_params_schema(opener, data_id)

    def open_data(self,
                  data_id: str,
                  opener_id: str = None,
                  **open_params) -> xr.Dataset:
        opener = self._find_opener(opener_id=opener_id, data_id=data_id)
        open_params_schema = self._get_open_data_params_schema(opener,
                                                               data_id)
        assert_valid_params(open_params,
                            name='open_params',
                            schema=open_params_schema)
        fs_path = self._convert_data_id_into_fs_path(data_id)
        return opener.open_data(fs_path,
                                fs=self.fs,
                                root=self.root,
                                **open_params)

    def get_data_writer_ids(self, data_type: str = None) \
            -> Tuple[str, ...]:
        data_type = DataType.normalize(data_type)
        return tuple(ext.name for ext in find_data_writer_extensions(
            predicate=get_data_accessor_predicate(
                data_type=data_type,
                storage_id=self.protocol
            )
        ))

    def get_write_data_params_schema(self, writer_id: str = None) \
            -> JsonObjectSchema:
        writer = self._find_writer(writer_id=writer_id)
        return self._get_write_data_params_schema(writer)

    def write_data(self,
                   data: Any,
                   data_id: str = None,
                   writer_id: str = None,
                   replace: bool = False,
                   **write_params) -> str:
        if self.read_only:
            raise DataStoreError('Data store is read-only.')
        if not writer_id:
            writer_id = self._guess_writer_id(data, data_id=data_id)
        writer = self._find_writer(writer_id=writer_id)
        write_params_schema = self._get_write_data_params_schema(writer)
        assert_valid_params(write_params,
                            name='write_params',
                            schema=write_params_schema)
        data_id = self._ensure_valid_data_id(writer_id, data_id=data_id)
        fs_path = self._convert_data_id_into_fs_path(data_id)
        self.fs.makedirs(self.root, exist_ok=True)
        written_fs_path = writer.write_data(data,
                                            fs_path,
                                            replace=replace,
                                            fs=self.fs,
                                            root=self.root,
                                            **write_params)
        # Verify, accessors fulfill their write_data() contract
        assert_true(fs_path == written_fs_path,
                    message='FsDataAccessor implementations must '
                            'return the data_id passed in.')
        # Return original data_id (which is a relative path).
        # Note: it would be cleaner to return written_fs_path
        # here, but it is an absolute path.
        # --> Possible solution: FsDataAccessor implementations
        # should be responsible for resolving their data_id into a
        # filesystem path by providing them both with fs and root
        # arguments.
        return data_id

    def get_delete_data_params_schema(self, data_id: str = None) \
            -> JsonObjectSchema:
        writer = self._find_writer(data_id=data_id)
        return self._get_delete_data_params_schema(writer, data_id)

    def delete_data(self, data_id: str, **delete_params):
        if self.read_only:
            raise DataStoreError('Data store is read-only.')
        writer = self._find_writer(data_id=data_id)
        delete_params_schema = self._get_delete_data_params_schema(writer,
                                                                   data_id)
        assert_valid_params(delete_params,
                            name='delete_params',
                            schema=delete_params_schema)
        fs_path = self._convert_data_id_into_fs_path(data_id)
        writer.delete_data(fs_path,
                           fs=self.fs,
                           root=self.root,
                           **delete_params)

    def register_data(self, data_id: str, data: Any):
        # We don't need this as we use the filesystem
        pass

    def deregister_data(self, data_id: str):
        # We don't need this as we use the filesystem
        pass

    ###############################################################
    # Implementation helpers

    @staticmethod
    def _get_open_data_params_schema(opener: DataOpener,
                                     data_id: str):
        schema = opener.get_open_data_params_schema(data_id=data_id)
        return FsAccessor.remove_storage_options_from_params_schema(schema)

    @staticmethod
    def _get_write_data_params_schema(writer: DataWriter):
        schema = writer.get_write_data_params_schema()
        return FsAccessor.remove_storage_options_from_params_schema(schema)

    @staticmethod
    def _get_delete_data_params_schema(writer: DataWriter,
                                       data_id: str):
        schema = writer.get_delete_data_params_schema(data_id)
        return FsAccessor.remove_storage_options_from_params_schema(schema)

    def _guess_writer_id(self, data, data_id: str = None):
        data_type = None
        format_id = None
        if data_id:
            accessor_id_parts = self._guess_accessor_id_parts(
                data_id, require=False
            )
            if accessor_id_parts:
                data_type = accessor_id_parts[0]
                format_id = accessor_id_parts[1]
        if isinstance(data, xr.Dataset):
            data_type = DATASET_TYPE.alias
            format_id = format_id or 'zarr'
        elif isinstance(data, MultiLevelDataset):
            data_type = MULTI_LEVEL_DATASET_TYPE.alias
            format_id = format_id or 'levels'
        elif isinstance(data, gpd.GeoDataFrame):
            data_type = GEO_DATA_FRAME_TYPE.alias
            format_id = format_id or 'geojson'
        predicate = get_data_accessor_predicate(
            data_type=data_type,
            format_id=format_id,
            storage_id=self.protocol
        )
        extensions = find_data_writer_extensions(predicate=predicate)
        if not extensions:
            raise DataStoreError(f'Can not find suitable data writer'
                                 f' for data of type {type(data)!r}'
                                 f' and format {format_id!r}')
        return extensions[0].name

    def _find_opener(self,
                     opener_id: str = None,
                     data_id: str = None,
                     require: bool = True) -> Optional[DataOpener]:
        if not opener_id:
            opener_id = self._find_opener_id(data_id=data_id, require=require)
            if opener_id is None:
                return None
        return new_data_opener(opener_id)

    def _find_writer(self,
                     writer_id: str = None,
                     data_id: str = None,
                     require: bool = True) -> Optional[DataWriter]:
        if not writer_id:
            writer_id = self._find_writer_id(data_id=data_id, require=require)
            if writer_id is None:
                return None
        return new_data_writer(writer_id)

    def _is_data_specified(self,
                           data_id: str,
                           data_type: DataTypeLike,
                           require: bool = False) -> bool:
        data_type = DataType.normalize(data_type)
        actual_data_type = self._guess_data_type_for_data_id(
            data_id, require=False
        )
        if actual_data_type is None:
            if require:
                raise DataStoreError(f'Cannot determine data type of'
                                     f' resource {data_id!r}')
            return False
        if not data_type.is_super_type_of(actual_data_type):
            if require:
                raise DataStoreError(f'Data type {data_type!r}'
                                     f' is not compatible with type'
                                     f' {actual_data_type!r} of'
                                     f' data resource {data_id!r}')
            return False
        return True

    def _assert_data_specified(self, data_id, data_type: DataTypeLike):
        self._is_data_specified(data_id, data_type, require=True)

    @classmethod
    def _ensure_valid_data_id(cls, writer_id: str, data_id: str = None) -> str:
        format_id = writer_id.split(':')[1]
        first_ext = None
        for known_ext, known_format_id in _FILENAME_EXT_TO_FORMAT.items():
            # Note, there may be multiple common file extensions
            # for a given data format_id, e.g. .tif, .tiff, .geotiff.
            # Must try them all:
            if format_id == known_format_id:
                if first_ext is None:
                    first_ext = known_ext
                if data_id and data_id.endswith(known_ext):
                    return data_id
        assert first_ext is not None
        if data_id:
            warnings.warn(f'Data resource identifier {data_id!r} is'
                          f' lacking an expected extension {first_ext!r}.'
                          f' It will be written as {format_id!r},'
                          f' but the store may later have difficulties'
                          f' identifying the correct data format.')
            return data_id
        return str(uuid.uuid4()) + first_ext

    def _assert_valid_data_id(self, data_id):
        if not self.has_data(data_id):
            raise DataStoreError(f'Data resource "{data_id}"'
                                 f' does not exist in store')

    def _convert_data_id_into_fs_path(self, data_id: str) -> str:
        assert_given(data_id, 'data_id')
        root = self.root
        fs_path = f'{root}/{data_id}' if root else data_id
        return fs_path

    def _assert_valid_data_type(self,
                                data_type: DataType):
        if data_type != ANY_TYPE:
            assert_in(data_type,
                      self.get_data_types(),
                      name='data_type')

    def _find_opener_id(self, data_id: str = None, require=True) \
            -> Optional[str]:
        return self._find_accessor_id(find_data_opener_extensions,
                                      data_id=data_id,
                                      require=require)

    def _find_writer_id(self, data_id: str = None, require=True) \
            -> Optional[str]:
        return self._find_accessor_id(find_data_writer_extensions,
                                      data_id=data_id,
                                      require=require)

    def _find_accessor_id(self,
                          find_data_accessor_extensions: Callable,
                          data_id: str = None,
                          require=True) -> Optional[str]:
        extensions = self._find_accessor_extensions(
            find_data_accessor_extensions,
            data_id=data_id,
            require=require
        )
        return extensions[0].name if extensions else None

    def _find_opener_extensions(self, data_id: str = None, require=True):
        return self._find_accessor_extensions(find_data_opener_extensions,
                                              data_id=data_id,
                                              require=require)

    def _find_writer_extensions(self, data_id: str = None, require=True):
        return self._find_accessor_extensions(find_data_writer_extensions,
                                              data_id=data_id,
                                              require=require)

    def _find_accessor_extensions(self,
                                  find_data_accessor_extensions: Callable,
                                  data_id: str = None,
                                  require=True) -> List[Extension]:
        if data_id:
            accessor_id_parts = self._guess_accessor_id_parts(
                data_id,
                require=require
            )
            if not accessor_id_parts:
                return []
            data_type_alias = accessor_id_parts[0]
            format_id = accessor_id_parts[1]
            storage_id = accessor_id_parts[2]
        else:
            data_type_alias = _DEFAULT_DATA_TYPE
            format_id = _DEFAULT_FORMAT_ID
            storage_id = self.protocol
        predicate = get_data_accessor_predicate(
            data_type=data_type_alias,
            format_id=format_id,
            storage_id=storage_id
        )
        extensions = find_data_accessor_extensions(predicate)
        if not extensions:
            if require:
                msg = 'No data accessor found'
                if data_id:
                    msg += f' for data resource {data_id!r}'
                raise DataStoreError(msg)
            return []
        return extensions

    def _guess_data_type_for_data_id(self, data_id: str, require=True) \
            -> Optional[DataType]:
        accessor_id_parts = self._guess_accessor_id_parts(data_id,
                                                          require=require)
        if accessor_id_parts is None:
            return None
        data_type_alias, _, _ = accessor_id_parts
        return DataType.normalize(data_type_alias)

    def _guess_accessor_id_parts(self, data_id: str, require=True) \
            -> Optional[Tuple[str, str, str]]:
        assert_given(data_id, 'data_id')
        ext = self._get_filename_ext(data_id)
        data_type_aliases = None
        format_id = _FILENAME_EXT_TO_FORMAT.get(ext.lower())
        if format_id is not None:
            data_type_aliases = _FORMAT_TO_DATA_TYPE_ALIASES.get(format_id)
        if data_type_aliases is None or format_id is None:
            if require:
                raise DataStoreError(f'Cannot determine data type for '
                                     f' data resource {data_id!r}')
            return None
        return data_type_aliases[0], format_id, self.protocol

    def _get_filename_ext(self, data_path: str) -> str:
        dot_pos = data_path.rfind('.')
        if dot_pos == -1:
            return ''
        slash_pos = data_path.rfind('/')
        if dot_pos < slash_pos:
            return ''
        # TODO: avoid self.fs here to ensure laziness.
        if self.fs.sep != '/':
            sep_pos = data_path.rfind(self.fs.sep)
            if dot_pos < sep_pos:
                return ''
        return data_path[dot_pos:]

    def _generate_data_ids(self,
                           dir_path: str,
                           data_type: DataType,
                           return_tuples: bool,
                           current_depth: int) -> _DataIds:
        root = self.root + ('/' + dir_path if dir_path else '')
        if not self.fs.exists(root):
            return
        # noinspection PyArgumentEqualDefault
        for file_info in self.fs.ls(root, detail=True):
            file_path: str = file_info['name']
            if file_path.startswith(self.root):
                file_path = file_path[len(self.root) + 1:]
            elif file_path.startswith('/' + self.root):
                file_path = file_path[len(self.root) + 2:]
            if not file_path:
                continue
            if self._is_data_specified(file_path, data_type):
                yield (file_path, {}) if return_tuples else file_path
            elif file_info.get('type') == 'directory' \
                    and (self._max_depth is None
                         or current_depth < self._max_depth):
                yield from self._generate_data_ids(file_path,
                                                   data_type,
                                                   return_tuples,
                                                   current_depth + 1)

    def _filter_data_ids(self,
                         data_ids: _DataIds,
                         return_tuples: bool) -> _DataIds:
        for data_id in data_ids:
            path = data_id[0] if return_tuples else data_id
            excluded = False
            for pattern in self._excludes:
                excluded = fnmatch.fnmatch(path, pattern)
                if excluded:
                    break
            if not excluded:
                included = True
                for pattern in self._includes:
                    included = fnmatch.fnmatch(path, pattern)
                    if included:
                        break
                if included:
                    yield data_id

    @staticmethod
    def _normalize_wc(wc: Optional[Union[str, Sequence[str]]]) -> Tuple[str]:
        return tuple() if not wc \
            else (wc,) if isinstance(wc, str) \
            else tuple(wc)


class FsDataStore(BaseFsDataStore, FsAccessor):
    """
    Specialization of a :class:BaseFsDataStore that
    also implements a :class:FsAccessor which serves
    the filesystem.

    The data store is capable of filtering the data identifiers reported
    by ``get_data_ids()``. For this purpose the optional keywords
    `excludes` and `includes` are used which can both take the form of
    a wildcard pattern or a sequence of wildcard patterns:

    * ``excludes``: if given and if any pattern matches the identifier,
      the identifier is not reported.
    * ``includes``: if not given or if any pattern matches the identifier,
      the identifier is reported.

    :param root: Root or base directory.
        Defaults to "".
    :param max_depth: Maximum recursion depth. None means limitless.
        Defaults to 1.
    :param read_only: Whether this is a read-only store.
        Defaults to False.
    :param includes: Optional sequence of wildcards that include
        certain filesystem paths. Affects the data identifiers (paths)
        returned by `get_data_ids()`. By default, all paths are included.
    :param excludes: Optional sequence of wildcards that exclude
        certain filesystem paths. Affects the data identifiers (paths)
        returned by `get_data_ids()`. By default, no paths are excluded.
    :param storage_options: Parameters specific to the underlying filesystem.
        Used to instantiate the filesystem.
    """

    def __init__(self,
                 root: str = '',
                 max_depth: Optional[int] = 1,
                 read_only: bool = False,
                 includes: Optional[Sequence[str]] = None,
                 excludes: Optional[Sequence[str]] = None,
                 storage_options: Dict[str, Any] = None):
        self._storage_options = storage_options or {}
        super().__init__(root=root,
                         max_depth=max_depth,
                         read_only=read_only,
                         includes=includes,
                         excludes=excludes)

    @property
    def protocol(self) -> str:
        # Base class returns self.fs.protocol which
        # instantiates the filesystem.
        # Avoid this, as we know the protocol up front.
        return self.get_protocol()

    @property
    def storage_options(self) -> Dict[str, Any]:
        return self._storage_options

    def _load_fs(self) -> fsspec.AbstractFileSystem:
        # Note, this is invoked only once per store instance.
        fs, _, _ = self.load_fs(
            {STORAGE_OPTIONS_PARAM_NAME: self._storage_options}
        )
        return fs

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        return cls.add_storage_options_to_params_schema(
            super().get_data_store_params_schema()
        )



import os
from typing import Set, Mapping, Tuple, Union

import xarray as xr


def open_sentinel3_product(path: str,
                           var_names: Set[str] = None,
                           chunks: Mapping[str, Union[int, Tuple[int, ...]]] = None) -> xr.Dataset:
    """
    Open a Sentinel-3 product from given path.

    :param chunks:
    :param path: Sentinel-3 product path
    :param var_names: Optional variable names to be included.
    :param chunks: Optional mapping from dimension name to chunk sizes or chunk size tuples.
    :return: A dataset representation of the Sentinel-3 product.
    """
    x_name = 'longitude'
    y_name = 'latitude'
    data_vars = {}
    geo_vars_file_name = 'geo_coordinates.nc'
    file_names = set(file_name for file_name in os.listdir(path) if file_name.endswith('.nc'))
    if geo_vars_file_name not in file_names:
        raise ValueError(f'missing file {geo_vars_file_name!r} in {path}')
    file_names.remove(geo_vars_file_name)
    geo_vars_path = os.path.join(path, geo_vars_file_name)
    with _open_dataset(geo_vars_path, chunks=chunks) as geo_ds:
        if x_name not in geo_ds:
            raise ValueError(f'variable {x_name!r} not found in {geo_vars_path}')
        if y_name not in geo_ds:
            raise ValueError(f'variable {y_name!r} not found in {geo_vars_path}')
        x_var = geo_ds[x_name]
        y_var = geo_ds[y_name]
        if x_var.ndim != 2:
            raise ValueError(f'variable {x_name!r} must have two dimensions')
        if y_var.ndim != x_var.ndim \
                or y_var.shape != x_var.shape \
                or y_var.dims != x_var.dims:
            raise ValueError(f'variable {y_name!r} must have same shape and dimensions as {x_name!r}')
        data_vars.update({x_name: x_var, y_name: y_var})
    for file_name in file_names:
        dataset_path = os.path.join(path, file_name)
        with _open_dataset(dataset_path, chunks=chunks) as ds:
            for var_name, var in ds.data_vars.items():
                if var_names and var_name not in var_names:
                    continue
                if var.ndim >= 2 \
                        and var.shape[-2:] == x_var.shape \
                        and var.dims[-2:] == x_var.dims:
                    data_vars.update({var_name: var})
    return xr.Dataset(data_vars)


def _open_dataset(dataset_path, chunks=None) -> xr.Dataset:
    if chunks is None:
        ds = xr.open_dataset(dataset_path)
        for var_name, var in ds.variables.items():
            chunk_sizes = var.encoding.get('chunksizes')
            if isinstance(chunk_sizes, (tuple, list)) and len(chunk_sizes) == len(var.dims):
                if chunks is None:
                    chunks = dict()
                chunks.update({dim: size for dim, size in zip(var.dims, chunk_sizes)})
        if chunks is None:
            return ds
    return xr.open_dataset(dataset_path, chunks=chunks)


def is_sentinel3_product(path: str) -> bool:
    """
    Test if given *path* is likely a Sentinel-3 product path.

    :param path: (directory) path
    :return: True, if so
    """
    return os.path.isdir(path) and os.path.isfile(os.path.join(path, 'geo_coordinates.nc'))
